/*-
 * Copyright 2007 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <limits.h>
#include <pthread.h>
#include <sys/errno.h>
#include <string>
#include <iostream>

using namespace std;

#include "cx88_xml_config.h"
#include "cx88_file_capture.h"
#include "cx88_udp_capture.h"
#include "cx88_mpeg_driver_factory.h"

#define CX88_DEFAULT_UDP_PORT 8802
#define DEFAULT_XML_CONFIG "~/.cx88.xml"
#define DEFAULT_TUNER_CONFIG "~/.tuner_config"

extern char *optarg;

typedef struct
{
   cx88_profile_t profile;
   unsigned long channel;
} cx88_channel;

typedef struct
{
   pthread_mutex_t lock;
   pthread_cond_t signal;
} cx88_completion_arg;

int parse_channel(cx88_xml_config *xmlconf, string &channel, cx88_channel &dest);

static int noninteractive_capture(
   cx88_driver *driver,
   cx88_channel &channel,
   unsigned int duration_s);

static int interactive_capture(
   cx88_driver *driver,
   cx88_channel &channel,
   cx88_xml_config *xmlconf);
   
static void capture_complete(void *arg);

int main(int argc, char **argv)
{
   int error = 0, optch = -1;
   char *device = NULL;
   cx88_driver *driver = NULL;
   cx88_xml_config *xmlconf = NULL;
   tuner_config *tunerconf = NULL;
   capvec captures;
   cx88_channel init_channel;
   init_channel.profile = NULL;
   string channel = "";
   unsigned int duration_s = 0;
   static const char *help = 
      "Options:\n"
      "\t-d <path>: [Required] Specify device from which to capture\n"
      "\t-c <channel | profile:channel>: [Required] Specify initial channel, e.g. '7' or 'ATSC:7'\n"
      "\t-u <host-ip | host-ip:port>: [Optional] Send capture over UDP to provided host\n"
      "\t-f <filename>: [Optional] Send capture to specified file\n"
      "\t-x <filename>: [Optional] Specify XML configuration file; default is "DEFAULT_XML_CONFIG"\n"
      "\t-t <filename>: [Optional] Specify libtuner configuration file; default is "DEFAULT_TUNER_CONFIG"\n"
      "\t-n <value>: [Optional] Specify non-interactive capture for provided number of minutes\n"
      "\t-? : Display this message\n";
   while (!error && ((optch = getopt(argc, argv, "d:c:u:f:x:t:n:?")) != -1))
   {
      switch (optch) 
      {
         case 'd':
            if (device != NULL)
            {
               CX88ERR << "Warning: ignoring duplicate device directive" << endl;
            }
            else if ((device = strdup(optarg)) == NULL)
            {
               CX88ERR << "Unable to store device filename." << endl;
               error = ENOMEM;
            }
            break;
         case 'c':
            if (channel.length() > 0)
            {
               CX88ERR << "Warning: ignoring duplicate channel directive" << endl;
               break;
            }
            try
            {
               channel = optarg;
            }
            catch (...)
            {
               CX88ERR << "Unable to store channel name.";
               error = ENOMEM;
            }
            break;
         case 'u':
         {
            char *hostname = strdup(optarg);
            unsigned short portnum = CX88_DEFAULT_UDP_PORT;
            if (hostname == NULL)
            {
               CX88ERR << "Unable to store UDP target address." << endl;
               error = ENOMEM;
               break;
            }      
            char *port = strchr(hostname, ':');
            if (port != NULL)
            {
               port[0] = '\0';
               portnum = (unsigned short) (strtoul(port + 1, NULL, 10));
            }
            cx88_udp_capture *cap = new cx88_udp_capture(hostname, portnum, error);
            if (cap == NULL)
            {
               CX88ERR << "Unable to store UDP capture object." << endl;
               error = ENOMEM;
            }
            try
            {
               captures.push_back(cap);    
            }
            catch (...)
            {
               delete cap;
               CX88ERR << "Unable to store UDP capture object." << endl;
               error = ENOMEM;
            }
            free(hostname);
            break;
         }
         case 'f':
         {
            cx88_file_capture *cap = new cx88_file_capture(optarg, error);
            if (cap == NULL)
            {
               CX88ERR << "Unable to store file capture object." << endl;
               error = ENOMEM;
            }
            try
            {
               captures.push_back(cap);    
            }
            catch (...)
            {
               delete cap;
               CX88ERR << "Unable to store file capture object." << endl;
               error = ENOMEM;
            }
            break;
         }
         case 'x':
            xmlconf = new cx88_xml_config(optarg, error);
            if (xmlconf == NULL)
            {
               CX88ERR << "Unable to store XML configuration object." << endl;
               error = ENOMEM;
            }
            if (error)
            {
               delete xmlconf;
               xmlconf = NULL;
            }
            break;
         case 't':
            tunerconf = new tuner_config(optarg, error);
            if (tunerconf == NULL)
            {
               CX88ERR << "Unable to store libtuner configuration object." << endl;
               error = ENOMEM;
            }
            if (error)
            {
               delete tunerconf;
               tunerconf = NULL;
            }
            break;
         case 'n':
            duration_s = (unsigned long) (strtod(optarg, NULL) * 60);
            break;
         case '?':
            cout << help;
            break;
         default:
            break;
      }
   }
   if (!error)
   {
      if (device == NULL)
      {
         CX88ERR << "Error: You must specify a device (-d)." << endl;
         error = EINVAL;
      }
      if (channel.length() == 0)
      {
         CX88ERR << "Error: You must specify an initial channel (-c)." << endl;
         error = EINVAL;
      }
      if (captures.size() == 0)
      {
         CX88ERR << "Error: You must specify at least one capture (-u, -f)." << endl;
         error = EINVAL;
      }
   }
   if (!error && (xmlconf == NULL))
   {
      xmlconf = new cx88_xml_config(DEFAULT_XML_CONFIG, error);
      if (xmlconf == NULL)
      {  
         error = ENOMEM;  
      }
      if (error)
      {
         CX88ERR << "Unable to load cx88 XML configuration file" << endl;
      }
   }
   if (!error && (tunerconf == NULL))
   {
      tunerconf = new tuner_config(DEFAULT_TUNER_CONFIG, error);
      if (tunerconf == NULL)
      {
         error = ENOMEM;
      }
      if (error)
      {
         CX88ERR << "Unable to load libtuner configuration file" << endl;
      }
   }
   if (!error)
   {
      if (strstr(device, "cx88mpeg") != NULL)
      {
         driver = cx88_mpeg_driver_factory::create(
            *xmlconf,
            *tunerconf,
            device,
            captures,
            error);
         if (!error && (driver == NULL))
         {
            error = ENOMEM;  
         }
      }
   }
   if (!error)
   {
      error = parse_channel(xmlconf, channel, init_channel);
   }
   if (!error)
   {
      if (duration_s == 0)
      {
         error = interactive_capture(driver, init_channel, xmlconf);
      }
      else
      {
         error = noninteractive_capture(driver, init_channel, duration_s);  
      }
   }
   for (capvec::size_type i = 0; i < captures.size(); ++i)
   {
      delete captures[i];  
   }
   delete driver;
   free(device);
   delete xmlconf;
   delete tunerconf;
   return error;
}

int parse_channel(cx88_xml_config *xmlconf, string &channel, cx88_channel &dest)
{
   int error = 0;
   string::size_type chan_begin = channel.find_first_of(':', 0);
   string chanstr = channel;
   string profile = channel;
   const char *profile_name = NULL;
   if (chan_begin != string::npos)
   {
      chanstr = channel.substr(chan_begin + 1, channel.length() - chan_begin - 1);
      profile = channel.substr(0, chan_begin);
      profile_name = profile.c_str();
   }
   if ((profile_name != NULL) || (dest.profile == NULL))
   {
      if ((error = xmlconf->get_profile(profile_name, dest.profile)))
      {
         return error;
      }
   }
   dest.channel = strtoul(chanstr.c_str(), NULL, 10);
   return error;
}

static void capture_complete(void *arg)
{
   cx88_completion_arg *completion = reinterpret_cast <cx88_completion_arg*> (arg);
   pthread_mutex_lock(&(completion->lock));
   pthread_cond_signal(&(completion->signal));
   pthread_mutex_unlock(&(completion->lock));   
}

static int noninteractive_capture(
   cx88_driver *driver,
   cx88_channel &channel,
   unsigned int duration_s)
{
   int error = 0;
   cx88_completion_arg arg;
   cx88_capture_params params;
   params.completion = capture_complete;
   params.completion_arg = &arg;
   params.duration_s = duration_s;
   if ((error = pthread_mutex_init(&(arg.lock), NULL)))
   {
      return error;
   }
   if ((error = pthread_cond_init(&(arg.signal), NULL)))
   {
      pthread_mutex_destroy(&(arg.lock));
      return error;
   }
   pthread_mutex_lock(&(arg.lock));
   if ((error = driver->start(channel.profile, channel.channel, params)))
   {
      pthread_mutex_unlock(&(arg.lock));
      pthread_cond_destroy(&(arg.signal));
      pthread_mutex_destroy(&(arg.lock));
      return error;
   }
   pthread_cond_wait(&(arg.signal), &(arg.lock));
   pthread_mutex_unlock(&(arg.lock));
   driver->stop();
   pthread_cond_destroy(&(arg.signal));
   pthread_mutex_destroy(&(arg.lock));
   return error;
}

static int interactive_capture(
   cx88_driver *driver,
   cx88_channel &channel,
   cx88_xml_config *xmlconf)
{
   int error = 0;
   string str;
   cx88_capture_params params;
   params.completion = NULL;
   params.completion_arg = NULL;
   params.duration_s = 0;
   if ((error = driver->start(channel.profile, channel.channel, params)))
   {
      return error;
   }
   int cmd;
   while ((cmd = getchar()) != 'q')
   {
      switch (cmd)
      {
         case 'c':
            cout << "Enter channel:" << endl;
            getline(cin, str);
            getline(cin, str);
            if (parse_channel(xmlconf, str, channel) != 0)
            {
               CX88ERR << "Invalid channel identifier" << endl;
            }
            else if (driver->set_channel(channel.profile, channel.channel) != 0)
            {
               CX88ERR << "Unable to change channel" << endl;
            }
            break;
         case 'i':
            if (driver->get_info(str) != 0)
            {
               CX88ERR << "Unable to get driver info" << endl;  
            }
            else
            {
               cout << str.c_str() << endl;
            }
            break;
         default:
            break;
      }
   }
   driver->stop();
   return error;
}
