/*-
 * Copyright 2007 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <sys/errno.h>
#include "cx88_driver.h"

cx88_driver::cx88_driver(
   cx88_xml_config &cx88conf,
   tuner_config &tunerconf,
   const char *devnode,
   capvec &captures)
   : m_cx88conf(cx88conf),
     m_tunerconf(tunerconf),
     m_devnode(devnode),
     m_stop(false),
     m_captures(captures),
     m_state(CX88_DRIVER_STOPPED),
     m_profile(NULL),
     m_channel(0)
{  
}

int cx88_driver::set_channel(const cx88_profile_t &profile, unsigned long channel)
{
   if (m_state == CX88_DRIVER_RUNNING)
   {
      return change_channel(profile, channel);
   }
   else if (m_state == CX88_DRIVER_PAUSED)
   {
      return start(profile, channel, m_params);
   }
   else
   {
      return ENXIO;
   }
}

int cx88_driver::start(const cx88_profile_t &profile, unsigned long channel, cx88_capture_params &params)
{
   int error = 0;
   if (m_state != CX88_DRIVER_RUNNING)
   {
      m_profile = profile;
      m_channel = channel;
      m_params = params;
      m_state = CX88_DRIVER_RUNNING;
      if ((error = pthread_create(&m_thread, NULL, thread_routine, this)))
      {
         m_state = CX88_DRIVER_STOPPED;
         return error;
      }
   }
   return error;
}

int cx88_driver::stop(void)
{
   void *retval = NULL;
   if (m_state == CX88_DRIVER_RUNNING)
   {
      m_stop = true;
      pthread_join(m_thread, &retval);
      m_stop = false;
      m_state = CX88_DRIVER_STOPPED;
   }
   return ((int) retval);
}

void *cx88_driver::thread_routine(void *arg)
{
   cx88_driver *driver = reinterpret_cast<cx88_driver*>(arg);
   int error = driver->capture(driver->m_profile, driver->m_channel, driver->m_params);
   driver->m_state = CX88_DRIVER_PAUSED;
   return ((void*)error);
}

int cx88_driver::start_captures(void)
{
   int error = 0;
   for (capvec::size_type i = 0; i < m_captures.size(); ++i)
   {
      if ((error = m_captures[i]->start()))
      {
         break;
      }
   }
   return error;
}

int cx88_driver::submit_buffer(void *buffer, size_t num_bytes, size_t packet_size)
{
   int error = 0;
   for (capvec::size_type i = 0; i < m_captures.size(); ++i)
   {
      if ((error = m_captures[i]->submit_buffer(buffer, num_bytes, packet_size)))
      {
         break;
      }
   }
   return error;
}

void cx88_driver::stop_captures(void)
{
   for (capvec::size_type i = 0; i < m_captures.size(); ++i)
   {
      m_captures[i]->stop();
   }
}
