/*-
 * Copyright 2007 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <strings.h>
#include <sys/errno.h>
#include <sys/types.h>
#include <sys/mman.h>

#include <sstream>
using namespace std;

#include "cx88_mpeg_driver.h"

cx88_mpeg_driver::cx88_mpeg_driver(
   cx88_xml_config &cx88conf,
   tuner_config &tunerconf,
   const char *devnode,
   capvec &captures,
   int &error)
   : cx88_driver(cx88conf, tunerconf, devnode, captures),
     m_channel_changed(false),
     m_fd(-1),
     m_mutex_created(false)
{
   if (error)
   {
      return;
   }
   error = pthread_mutex_init(&m_devlock, NULL);
   if (!error)
   {
      m_mutex_created = true;
   }
}

cx88_mpeg_driver::~cx88_mpeg_driver(void)
{
   if (m_mutex_created)
   {
      pthread_mutex_destroy(&m_devlock);
      m_mutex_created = false;
   }
}

int cx88_mpeg_driver::capture(const cx88_profile_t &profile, unsigned long channel, cx88_capture_params &params)
{
   int error = 0;
   cx88_mpeg_config config;
   if ((m_fd = open(m_devnode, O_RDWR)) < 0)
   {
      return ENOENT;  
   }
   config.in_buftype = CX88_MPEG_BUF_KERNEL;
   config.in_packet_size = CX88_MPEG_DEFAULT_PACKET_SIZE;
   config.inout_num_packets = 10000;
   config.in_persist = 1;
   config.in_num_i2c_devices = 0;
   if ((error = set_config_options(config)))
   {
      close(m_fd);
      return error;  
   }
   if ((error = ioctl(m_fd, CX88_MPEG_CONFIG, &config)))
   {
      close(m_fd);
      return error;
   }
   void *buf = mmap(NULL, config.bufconfig.out_kernelbuf.out_mmap_size, PROT_READ, MAP_SHARED, m_fd, 0);
   if (buf == MAP_FAILED)
   {
      error = errno;
      close(m_fd);
      return error;
   }
   error = start_captures();
   if (!error)
   {
      error = i2c_setup(config);
   }
   if (!error)
   {
      error = change_channel(profile, channel);
   }
   if (!error)
   {
      error = capture(buf, config, params);  
   }
   i2c_teardown(config);
   stop_captures();
   munmap(buf, config.bufconfig.out_kernelbuf.out_mmap_size);
   close(m_fd);
   if (params.completion != NULL)
   {
      params.completion(params.completion_arg);  
   }
   return error;
}

int cx88_mpeg_driver::change_channel(cx88_mpeg_config &config, dvb_interface &ifc)
{
   int error = 0;
   m_channel_changed = false;
   for (uint8_t i = 0; i < config.in_num_i2c_devices; ++i)
   {
      if (m_i2c_devs[i].driver != NULL)
      {   
         if ((error = m_i2c_devs[i].driver->set_channel(m_channel, ifc)))
         {
            return error;
         }
      }
   }
   for (uint8_t i = 0; i < config.in_num_i2c_devices; ++i)
   {
      if (m_i2c_devs[i].driver != NULL)
      {   
         if ((error = m_i2c_devs[i].driver->start(5000)))
         {
            return error;
         }
      }
   }
   return error;
}

int cx88_mpeg_driver::capture(void *buffer, cx88_mpeg_config &config, cx88_capture_params &params)
{
   cx88_mpeg_capture capture;
   capture.in_start_byte = CX88_MPEG_DEFAULT_START_BYTE;
   capture.in_type = params.duration_s ? CX88_MPEG_CAP_DURATION : CX88_MPEG_CAP_CONTINUOUS;
   capture.in_restriction = params.duration_s;
   capture.in_options = 0;
   dvb_interface ifc;
   cx88_mpeg_wait wait;
   wait.in_timeout_s = 0;
   int error = 0;
   while (!m_stop)
   {
      pthread_mutex_lock(&m_devlock);
      error = set_channel_options();
      if (!error)
      {
         error = change_channel(config, ifc);
      }
      if (ifc.input_width_bits == DVB_INPUT_SERIAL)
      {
         capture.in_options |= CX88_MPEG_SERIAL_INPUT;
      }
      if (ifc.clock == DVB_IFC_PUNC_CLCK)
      {
         capture.in_options |= CX88_MPEG_PUNCTURED_CLOCK;
      }
      if (ifc.polarity == DVB_IFC_POS_POL)
      {
         capture.in_options |= CX88_MPEG_POS_POLARITY;
      }
      if (ifc.bit_endianness == DVB_IFC_BIT_LE)
      {
         capture.in_options |= CX88_MPEG_LITTLE_ENDIAN;
      }
      if (!error)
      {
         error = set_capture_options(capture);
      }
      pthread_mutex_unlock(&m_devlock);
      if (error || (error = ioctl(m_fd, CX88_MPEG_CAPTURE, &capture)))
      {
         return error;
      }
      uint8_t last_buffer = 0;
      while (!m_stop && !m_channel_changed)
      {
         if ((error = ioctl(m_fd, CX88_MPEG_WAIT, &wait)))
         {
            ioctl(m_fd, CX88_MPEG_STOP);
            return error;
         }
         if ((wait.out_status & CX88_MPEG_BUS_ERROR) == CX88_MPEG_BUS_ERROR)
         {
            CX88ERR << "Fatal error: PCI bus error" << endl;
         }
         if ((wait.out_status & CX88_MPEG_DMA_PROGRAM) == CX88_MPEG_DMA_PROGRAM)
         {
            CX88ERR << "Fatal error: DMA program error" << endl;
         }
         if ((wait.out_status & CX88_MPEG_WAIT_ABORTED) == CX88_MPEG_WAIT_ABORTED)
         {
            CX88ERR << "Fatal error: Wait aborted" << endl;
         }
         if ((wait.out_status & CX88_MPEG_CAP_TIMEOUT) == CX88_MPEG_CAP_TIMEOUT)
         {
            CX88ERR << "Nonfatal error: Capture timed out" << endl;
         }
         if ((wait.out_status & CX88_MPEG_BAD_PACKET) == CX88_MPEG_BAD_PACKET)
         {
            CX88ERR << "Nonfatal error: " << wait.out_num_bad_packets << " bad packet(s)" << endl;
         }
         if ((wait.out_status & CX88_MPEG_OVERFLOW) == CX88_MPEG_OVERFLOW)
         {
            CX88ERR << "Nonfatal error: " << wait.out_num_overflows << "overflow(s)" << endl;
         }
         if (wait.out_status & CX88_MPEG_CAP_FATAL)
         {
            ioctl(m_fd, CX88_MPEG_STOP);
            return EIO;
         }
         if (wait.out_buffer_available == 0 || wait.out_buffer_available > config.out_num_buffers)
         {
            ioctl(m_fd, CX88_MPEG_STOP);
            return EIO;
         }
         while (last_buffer != wait.out_buffer_available)
         {
            ++last_buffer;
            if (last_buffer > config.out_num_buffers)
            {
               last_buffer = 1;
            }
            size_t bufoffset = (last_buffer - 1) * config.bufconfig.out_kernelbuf.out_buffer_size;
            size_t bufsize = config.bufconfig.out_kernelbuf.out_mmap_size - bufoffset;
            void *bufptr = (char *)buffer + bufoffset;
            if (bufsize > (config.out_packets_per_buffer * config.in_packet_size))
            {
               bufsize = config.out_packets_per_buffer * config.in_packet_size;
            }
            submit_buffer(bufptr, bufsize, config.in_packet_size);
         }
         if (wait.out_transfer_complete)
         {
            return error;  
         }
      }
      ioctl(m_fd, CX88_MPEG_STOP);
   }
   return error;
}

int cx88_mpeg_driver::change_channel(const cx88_profile_t &profile, unsigned long channel)
{
   cx88_range_t range;
   int error = m_cx88conf.get_range(profile, channel, range);
   pthread_mutex_lock(&m_devlock);
   if (!error)
   {
      const char *bandwidth = m_cx88conf.get_property(range, CX88_XML_BANDWIDTH_NAME);
      const char *base_freq = m_cx88conf.get_property(range, CX88_XML_FREQ_NAME);
      const char *modulation = m_cx88conf.get_property(range, CX88_XML_MODULATION_NAME);
      if ((bandwidth != NULL) && (base_freq != NULL) && (modulation != NULL))
      {
         m_channel.bandwidth_hz = (uint32_t) (strtoul(bandwidth, NULL, 10));
         m_channel.frequency_hz = (uint32_t)
            (strtoul(base_freq, NULL, 10) + ((channel - range.start) * m_channel.bandwidth_hz));
         if (strcasecmp(modulation, "VSB8") == 0)
         {
            m_channel.modulation =  DVB_MOD_VSB_8; 
         }
         else if (strcasecmp(modulation, "VSB16") == 0)
         {
            m_channel.modulation = DVB_MOD_VSB_16;
         }
         else if (strcasecmp(modulation, "QAM64") == 0)
         {
            m_channel.modulation = DVB_MOD_QAM_64;
         }
         else if (strcasecmp(modulation, "QAM128") == 0)
         {
            m_channel.modulation = DVB_MOD_QAM_128;
         }
         else if (strcasecmp(modulation, "QAM256") == 0)
         {
            m_channel.modulation = DVB_MOD_QAM_256;
         }
         else if (strcasecmp(modulation, "QAM") == 0)
         {
            m_channel.modulation = DVB_MOD_QAM_AUTO;
         }
         else if (strcasecmp(modulation, "QPSK") == 0)
         {
            m_channel.modulation = DVB_MOD_QPSK;
         }
         else if (strcasecmp(modulation, "OFDM") == 0)
         {
            m_channel.modulation = DVB_MOD_OFDM;
         }
         else
         {
            error = EINVAL;
         }
      }
      else
      {
         error = EINVAL;
      }
   }
   m_channel_changed = true;
   pthread_mutex_unlock(&m_devlock);
   return error;
}

int cx88_mpeg_driver::get_info(string &info)
{
   int error = 0;
   pthread_mutex_lock(&m_devlock);
   try
   {
      dvb_signal signal;
      ostringstream stream;
      stream << "Tuned to " << m_channel.frequency_hz << "HZ" << endl;
      for (uint8_t i = 0; i < CX88_MPEG_MAX_I2C_DEVICES; ++i)
      {
         if (m_i2c_devs[i].driver != NULL)
         {   
            if ((error = m_i2c_devs[i].driver->get_signal(signal)))
            {
               pthread_mutex_unlock(&m_devlock);
               return error;
            }
         }
      }
      if (!(signal.locked))
      {
         stream << "\tNo signal" << endl;
      }
      else
      {
         stream << "\tSignal strength: " << signal.strength << "%\n\tSignal-to-noise ratio: " << signal.snr;
         stream << "\n\tBit-error rate: " << signal.ber << "\n\tUncorrected blocks: " << signal.uncorrected_blocks << endl;
      }
      info = stream.str();
   }
   catch(...)
   {
      error = ENOMEM;
   }
   pthread_mutex_unlock(&m_devlock);
   return 0;
}

int cx88_mpeg_driver::i2c_setup(cx88_mpeg_config &config)
{
   int error = 0;
   pthread_mutex_lock(&m_devlock);
   for (uint8_t i = 0; i < config.in_num_i2c_devices; ++i)
   {
      char i2c_devnode[CX88_MPEG_MAX_DEVNODE_NAME + 4];
      snprintf(i2c_devnode, CX88_MPEG_MAX_DEVNODE_NAME, "/dev/%s", config.out_i2c_devnodes[i]);
      m_i2c_devs[i].dev = new tuner_devnode_device(m_tunerconf, i2c_devnode, error);
      if (m_i2c_devs[i].dev == NULL) 
      {
         error = ENOMEM;
      }
      if (error)
      {
         break;  
      }
      m_i2c_devs[i].addr = config.in_i2c_addrs[i];
   }
   if (!error)
   {
      error = create_i2c_drivers();  
   }
   pthread_mutex_unlock(&m_devlock);
   return error;
}

void cx88_mpeg_driver::i2c_teardown(cx88_mpeg_config &config)
{
   pthread_mutex_lock(&m_devlock);
   destroy_i2c_drivers();
   for (uint8_t i = 0; i < config.in_num_i2c_devices; ++i)
   {
      if (m_i2c_devs[i].dev != NULL)
      {
         delete m_i2c_devs[i].dev;
         m_i2c_devs[i].dev = NULL;
         m_i2c_devs[i].driver = NULL;
      }
   }
   pthread_mutex_unlock(&m_devlock);
}
