/*-
 * Copyright 2007 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <sys/errno.h>
#include "cx88_xml_config.h"

cx88_xml_config::cx88_xml_config(const char *filename, int &error)
   : m_file(NULL)
{
   m_file = ezxml_parse_file(filename);
   if (m_file == NULL)
   {
      error = EINVAL;
   }
}

cx88_xml_config::~cx88_xml_config(void)
{
   if (m_file != NULL)
   {
      ezxml_free(m_file);
      m_file = NULL;  
   }
}

int cx88_xml_config::get_profile(const char *name, cx88_profile_t &profile)
{
   profile = NULL;
   ezxml_t curprofile = ezxml_child(m_file, CX88_XML_PROFILE_NAME);
   while (curprofile != NULL)
   {
      if (name == NULL)
      {
         profile = curprofile;
         return 0;  
      }
      const char *profile_name = ezxml_attr(curprofile, CX88_XML_NAME_NAME);
      if (strcmp(name, profile_name) == 0)
      {
         profile = curprofile;
         return 0;
      }
      curprofile = curprofile->next;  
   }
   return ENOENT;
}

int cx88_xml_config::get_range(const cx88_profile_t &profile, unsigned long channel, cx88_range_t &range)
{
   ezxml_t cur_range = ezxml_child(profile, CX88_XML_RANGE_NAME);
   while (cur_range != NULL)
   {
      unsigned long start_chan = 0, end_chan = 0;
      ezxml_t start = ezxml_child(cur_range, CX88_XML_START_NAME);
      ezxml_t end = ezxml_child(cur_range, CX88_XML_END_NAME);
      if (start != NULL)
      {
         start_chan = strtoul(start->txt, NULL, 10);
      }
      if (end != NULL)
      {
         end_chan = strtoul(end->txt, NULL, 10);
      }
      if ((channel >= start_chan) && (channel <= end_chan))
      {
         range.start = start_chan;
         range.end = end_chan;
         range.elem = cur_range;
         return 0;
      }
      cur_range = cur_range->next;
   }
   return ENOENT;
}

const char *cx88_xml_config::get_property(const cx88_range_t &range, const char *property)
{
   ezxml_t child = ezxml_child(range.elem, property);
   ezxml_t cur_level = range.elem;
   while (child == NULL)
   {
      cur_level = cur_level->parent;
      if (cur_level == NULL)
      {
         break;
      }
      child = ezxml_child(cur_level, property);
   }
   if (child != NULL)
   {
      return child->txt;
   }
   return NULL;
}
