/*-
 * Copyright 2007 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <sys/types.h>
#include <sys/systm.h>
#include <sys/param.h>
#include <sys/malloc.h>
#include <sys/kernel.h>
#include <sys/bus.h>
#include <sys/errno.h>
#include <sys/conf.h>
#include <sys/uio.h>
#include <machine/bus.h>
#include <machine/resource.h>
#include <sys/rman.h>
#include <sys/mman.h>
#include <sys/endian.h>
#include <sys/proc.h>
#include <sys/lock.h>
#include <sys/mutex.h>
#include <sys/condvar.h>
#include <dev/pci/pcivar.h>
#include <dev/pci/pcireg.h>
#include <vm/vm.h>
#include <vm/pmap.h>
#include <vm/vm_extern.h>
#include <sys/module.h>
#include "cx88_common.h"
#include "cx88_if.h"
#include "cx88_regs.h"

#define CX88_VID             0x14f1
#define CX88_BASE_PID        0x8800
#define CX88_DEFAULT_LATENCY 0x40
#define CX88_CHANNEL_OFFSET  0x340
#define CX88_USABLE_SRAM     0x7CC0
#define CX88_SRAM_OFFSET     0x180000
#define CX88_CDT_ENTRY_SIZE  0x10
#define CX88_MAX_NUM_FIFOS   0x3FF
#define CX88_MIN_INSTRQ_SIZE 0x40
#define CX88_MAX_INSTRQ_SIZE 0x3FC
#define CX88_REG_PTR1_BASE   0x300080
#define CX88_REG_PTR2_BASE   0x3000C0
#define CX88_REG_CNT1_BASE   0x300100
#define CX88_REG_CNT2_BASE   0x300140
#define CX88_REG_PTR1(chan)  (CX88_REG_PTR1_BASE + (chan * 0x4))
#define CX88_REG_PTR2(chan)  (CX88_REG_PTR2_BASE + (chan * 0x4))
#define CX88_REG_CNT1(chan)  (CX88_REG_CNT1_BASE + (chan * 0x4))
#define CX88_REG_CNT2(chan)  (CX88_REG_CNT2_BASE + (chan * 0x4))
#define CX88_CMDS(chan)      (CX88_SRAM_OFFSET + 0x40 + (chan * 0x40));
#define CX88_MAX_NSEGS       256
#define CX88_DMA_ALIGNMENT   4
#define CX88_DMA_BOUNDARY    0


static int cx88_release_softc(cx88_common_softc *sc);
static cx88_size_t cx88_fit_sram(cx88_size_t chunk_size, cx88_size_t used_sram);
static int cx88_alloc_channel(cx88_channel *channel);
static void cx88_dmaprogmap_load(
   void *arg,
   bus_dma_segment_t *sgl,
   int sgl_size,
   int error);
static void cx88_dmamap_load(
   void *arg,
   bus_dma_segment_t *sgl,
   int sgl_size,
   int error);
static void cx88_dmamap_load_uio(
   void *arg,
   bus_dma_segment_t *sgl,
   int sgl_size,
   bus_size_t mapsize,
   int error);
static int cx88_alloc_dmaprog_buffer(cx88_buffer *buffer);
static int cx88_alloc_dmaprogram(cx88_channel *channel);
static void cx88_release_dmaprogram(cx88_buffer *buffer);
static void cx88_dmamap_lock(void *arg, bus_dma_lock_op_t op);
static int cx88_configure_fifos(cx88_channel *channel);
static void *cx88_get_sibling_board_data(device_t dev);
static int cx88_init_board_data(void *arg);
static int cx88_destroy_board_data(void *arg);
static void cx88_init_dma_counters(cx88_buffer *buffer);

/* Channel descriptor table. min_size should sum to
 * CX88_USABLE_SRAM */
static const cx88_channel_desc channels[] =
{
   /*number               direction        min_size */
   /*Video*/
   {CX88_CH_VIDEO_PACKED, CX88_DOWNSTREAM, 0x2800},
   {CX88_CH_VIDEO_U,      CX88_DOWNSTREAM, 0x0800},
   {CX88_CH_VIDEO_V,      CX88_DOWNSTREAM, 0x0800},
   {CX88_CH_VIDEO_VBI,    CX88_DOWNSTREAM, 0x1400},
   /*Audio*/
   {CX88_CH_AUDIO_IN,     CX88_DOWNSTREAM, 0x0260},
   {CX88_CH_AUDIO_OUT,    CX88_UPSTREAM,   0x0260},
   /*Unused*/
   {CX88_CH_UNUSED,       CX88_UPSTREAM,   0x0000},
   /*MPEG*/
   {CX88_CH_MPEG_TS,      CX88_DOWNSTREAM, 0x1400},
   /*Video Interface Port*/
   {CX88_CH_VIP_IN,       CX88_DOWNSTREAM, 0x0600},
   {CX88_CH_VIP_OUT,      CX88_UPSTREAM,   0x0600},
   /*General Purpose I/O*/
   {CX88_CH_GPHOST_IN,    CX88_DOWNSTREAM, 0x0600},
   {CX88_CH_GPHOST_OUT,   CX88_UPSTREAM,   0x0600}
};

/* Static mutex for shared data initialization */
static struct mtx interdev_lock;

/* Module callback */
static int cx88_mod_handler(module_t mod, int what, void *arg)
{
   switch(what)
   {
      case MOD_LOAD:
         mtx_init(&interdev_lock, "CX88 Inter-device lock", NULL, MTX_DEF);
         return 0;
      case MOD_UNLOAD:
      case MOD_SHUTDOWN:
         mtx_lock(&interdev_lock);
         mtx_destroy(&interdev_lock);
         return 0;
      default:
         return 0;
   }
}

/* Module declarations */
static moduledata_t cx88_mod_data =
{
   "cx88",
   cx88_mod_handler,
   NULL
};

DECLARE_MODULE(cx88, cx88_mod_data, SI_SUB_DRIVERS, SI_ORDER_ANY);
MODULE_VERSION(cx88, CX88_MODVER);

/* Frees a buffer's RISC program */
static void cx88_release_dmaprogram(cx88_buffer *buffer)
{
   if (buffer->dma_program_buffer != NULL)
   {
      bus_dmamap_unload(buffer->channel->dma_program_tag, buffer->dma_program_map);
      bus_dmamap_destroy(buffer->channel->dma_program_tag, buffer->dma_program_map);
      contigfree(buffer->dma_program_buffer, buffer->dma_program_size, M_DEVBUF);
      buffer->dma_program_buffer = NULL;
      buffer->dma_program_size = 0;
   }
}

/* Free buffer resources */
void cx88_release_buffer(cx88_buffer *buffer)
{
   DIAGNOSTIC(device_printf(buffer->channel->sc->dev, "Releasing buffer for channel %d\n", buffer->channel->descriptor->number))
   if (buffer->data_buffer_type != CX88_DATABUF_NONE)
   {
      cx88_release_dmaprogram(buffer); /* Free the RISC program */
      bus_dmamap_unload(buffer->channel->data_buffer_tag, buffer->data_buffer_map);
      bus_dmamap_destroy(buffer->channel->data_buffer_tag, buffer->data_buffer_map);
      switch (buffer->data_buffer_type)
      {
         case CX88_DATABUF_MALLOC: /* Kernel-allocated buffer */
            DIAGNOSTIC(device_printf(buffer->channel->sc->dev, "Releasing kernel-allocated buffer\n"))
            free(buffer->data_buffer, M_DEVBUF);
            break;
         case CX88_DATABUF_USER: /* User-allocated buffer */
            DIAGNOSTIC(device_printf(buffer->channel->sc->dev, "Releasing user-allocated buffer\n"))
            vsunlock(buffer->data_buffer, buffer->data_buffer_size);
            break;
         default:
            break;
      }
      buffer->data_buffer = NULL;
      buffer->data_buffer_size = 0;
      if (buffer->sgl != NULL) /* Free scatter-gather list */
      {
         free(buffer->sgl, M_DEVBUF);
         buffer->sgl = NULL;
      }
      buffer->data_buffer_type = CX88_DATABUF_NONE;
      buffer->channel->is_dirty = TRUE;
   }
}

/* Free all buffers for a channel */
void cx88_release_buffers(cx88_channel *channel)
{
   if (channel->buffers != NULL)
   {
      uint8_t i;
      for (i = 0; i < channel->num_buffers; ++i)
      {
         cx88_release_buffer(&(channel->buffers[i]));
      }
      free(channel->buffers, M_DEVBUF);
      channel->buffers = NULL;
      channel->num_buffers = 0;
      channel->is_dirty = TRUE;
   }  
}

/* Free channel resources */
void cx88_release_channel(cx88_channel *channel)
{
   DIAGNOSTIC(device_printf(channel->sc->dev, "Releasing channel %d\n", channel->descriptor->number))
   cx88_release_buffers(channel); /* Free the buffers */
   if (channel->sram_region != NULL)
   {
      rman_release_resource(channel->sram_region);
      channel->sram_region = NULL;
   }
   channel->is_dirty = TRUE;
   channel->channel_size = 0;
   channel->max_line_size = 0;
   channel->packet_size = 0;
   channel->line_boundary = 0;
}

/* Initialize basic channel data */
int cx88_init_channel(
   cx88_channel *channel,
   cx88_common_softc *sc,
   cx88_channel_t channel_number,
   uint32_t max_buffer_size)
{
   DIAGNOSTIC(printf("Initializing channel %d\n", channel_number))
   if ((channel_number < 0) || (channel_number >= (sizeof(channels)) / sizeof(cx88_channel_desc))
      || (channels[channel_number].number != channel_number))
   {
      return EINVAL;
   }
   channel->descriptor = &(channels[channel_number]);
   channel->sc = sc;
   channel->buffers = NULL;
   channel->num_buffers = 0;
   channel->channel_offset = 0;
   channel->channel_size = 0;
   channel->packet_size = 0;
   channel->is_dirty = TRUE;
   channel->max_buffer_size = 0;
   channel->start_buffer = 0;
   channel->sram_region = NULL;
   channel->max_line_size = 0;
   channel->line_boundary = 0;
   if (max_buffer_size != 0)
   {
      int nsegments = max_buffer_size / PAGE_SIZE;
      sema_init(&(channel->map_semaphore), 0, "CX88 DMA map semaphore");
      int error = bus_dma_tag_create(
         bus_get_dma_tag(sc->dev),
         CX88_DMA_ALIGNMENT,
         CX88_DMA_BOUNDARY,
         BUS_SPACE_MAXADDR_32BIT,
         BUS_SPACE_MAXADDR,
         NULL,
         NULL,
         max_buffer_size,
         1,
         max_buffer_size,
         0,
         &cx88_dmamap_lock,
         &(channel->map_semaphore),
         &(channel->dma_program_tag));
      if (error)
      {
         device_printf(sc->dev, "Unable to create DMA program buffer tag\n");
         sema_destroy(&(channel->map_semaphore));
         return error;
      }
      if (max_buffer_size % PAGE_SIZE)
      {
         ++nsegments;
      }
      error = bus_dma_tag_create(
         bus_get_dma_tag(sc->dev),
         CX88_DMA_ALIGNMENT,
         CX88_DMA_BOUNDARY,
         BUS_SPACE_MAXADDR_32BIT,
         BUS_SPACE_MAXADDR,
         NULL,
         NULL,
         max_buffer_size,
         ((nsegments > CX88_MAX_NSEGS) ? CX88_MAX_NSEGS : nsegments),
         max_buffer_size,
         0,
         &cx88_dmamap_lock,
         &(channel->map_semaphore),
         &(channel->data_buffer_tag));
      if (error)
      {
         device_printf(sc->dev, "Unable to create DMA tag for buffer\n");
         bus_dma_tag_destroy(channel->dma_program_tag);
         sema_destroy(&(channel->map_semaphore));
         return error;
      }
      channel->max_buffer_size = max_buffer_size;
   }
   return 0;
}

/* Final destruction of channel */
void cx88_destroy_channel(cx88_channel *channel)
{
   if (channel->max_buffer_size != 0)
   {
      cx88_release_channel(channel); /* Free any allocated resources */
      bus_dma_tag_destroy(channel->dma_program_tag);
      bus_dma_tag_destroy(channel->data_buffer_tag);
      sema_destroy(&(channel->map_semaphore));
      channel->max_buffer_size = 0;
   }
}

/* Use 16.16 fixed-point math to determine the size ratio of a chunk of SRAM
 * to a larger region, and expand the chunk to occupy the same ratio of the entire
 * SRAM.  Used below to optimize channel allocation sizes when not all DMA
 * channels are used */
static cx88_size_t cx88_fit_sram(cx88_size_t chunk_size, cx88_size_t used_sram)
{
   uint32_t fixedpoint = (((uint32_t) chunk_size) << 16) / used_sram;
   fixedpoint *= CX88_USABLE_SRAM;
   cx88_size_t new_chunk_size = ((cx88_size_t) (fixedpoint >> 16));
   new_chunk_size -= (new_chunk_size % 8);
   return new_chunk_size;
}

/* Allocates a chunk of SRAM for a given DMA channel */
static int cx88_alloc_channel(cx88_channel *channel)
{
   DIAGNOSTIC(device_printf(channel->sc->dev, "Allocating channel number %d\n", channel->descriptor->number))
   uint16_t present_channels = channel->sc->board_data->present_channels;
   cx88_size_t total_size = 0, channel_size = 0, channel_offset = 0;
   int i;
   const size_t num_channels = (sizeof(channels) / sizeof(cx88_channel_desc));
   boolean_t realloc_attempted = FALSE;
   if (!(CX88_CHANNEL_MASK(channel->descriptor->number) & present_channels))
   {
      DIAGNOSTIC(device_printf(channel->sc->dev, "Channel %d not reserved by client\n", channel->descriptor->number))
      return EACCES;
   }
   /* Sum the min_size fields of all used channels for the board and
    * calculate the current channel's offset for the minimum allocation */
   for (i = 0; i < num_channels; ++i)
   {
      if (channel->descriptor->number == channels[i].number)
      {
         channel_offset = total_size;  
      }
      if (present_channels & CX88_CHANNEL_MASK(channels[i].number))
      {
         total_size += channels[i].min_size;
      }
   }
   /* Now expand the minimum size and offset calculations to fit the full SRAM size */
   channel_size = cx88_fit_sram(channel->descriptor->min_size, total_size);
   channel_offset = cx88_fit_sram(channel_offset, total_size);
   while (channel->sram_region == NULL)
   {
      // Step 1: try to allocate at channel_offset to help prevent fragging SRAM
      // if present_channels remains constant (the most likely case)
      channel->sram_region = rman_reserve_resource(&(channel->sc->board_data->sram),
         channel_offset, channel_offset + channel_size - 1, channel_size, 
         RF_ACTIVE | rman_make_alignment_flags(8), channel->sc->dev);
      if (channel->sram_region == NULL)
      {
         // Step 2: Now try to allocate anywhere
         channel->sram_region = rman_reserve_resource(&(channel->sc->board_data->sram),
            0, CX88_USABLE_SRAM - 1, channel_size, 
            RF_ACTIVE | rman_make_alignment_flags(8), channel->sc->dev);
      }
      if ((channel->sram_region == NULL) && (channel_size > channel->descriptor->min_size))
      {
         // Step 3: Try to allocate smallest allowable region
         channel->sram_region = rman_reserve_resource(&(channel->sc->board_data->sram),
            0, CX88_USABLE_SRAM - 1, channel->descriptor->min_size, 
            RF_ACTIVE | rman_make_alignment_flags(8), channel->sc->dev);
      }
      if (channel->sram_region == NULL)
      {
         // Step 4: See if siblings can reallocate
         device_printf(channel->sc->dev, 
            "Unable to allocate SRAM region for channel %d; reallocation request posted to siblings\n",
            channel->descriptor->number);
         // We haven't yet requested reallocation
         if (!realloc_attempted)
         {
            boolean_t retry = FALSE;
            int error = 0;
            for (i = 0; i < channel->sc->num_siblings; ++i)
            {
               error = CX88_REQUEST_RECONFIG(channel->sc->siblings[i]);
               if (error == 0)
               {
                  device_printf(channel->sc->siblings[i], "Reallocating\n");
                  /* If a sibling has reallocated, we should loop back and try
                   * to allocate our SRAM chunk again */
                  retry = TRUE;
               }
               else if (error == EINPROGRESS)
               {
                  device_printf(channel->sc->siblings[i], "Busy; will reallocate later\n");
               }
               else
               {
                  device_printf(channel->sc->siblings[i], "Unable to reallocate\n");
               }
            }
            if (!retry)
            {
               /* Break the loop if nobody has reallocated */
               return ENOMEM;
            }
            realloc_attempted = TRUE;
         }
         else
         {
            // We've already requested reallocation, so just give up.
            return ENOMEM;
         }
      }
   }
   channel->channel_offset = CX88_SRAM_OFFSET + CX88_CHANNEL_OFFSET + rman_get_start(channel->sram_region);
   channel->channel_size = rman_get_size(channel->sram_region);
   DIAGNOSTIC(
      device_printf(
         channel->sc->dev,
         "Allocated channel %d of size 0x%x at offset 0x%x\n",
         channel->descriptor->number,
         channel->channel_size,
         channel->channel_offset
         )
      )
   return 0;
}

/* Reset DMA program generation data */
static void cx88_init_dma_counters(cx88_buffer *buffer)
{
   buffer->dma_program_counter = 0;
   buffer->dma_byte_counter = 0;
   buffer->sgl_index = buffer->sgl;
   buffer->sgl_offset = 0;
}

/* Allocate FIFO space within a channel's SRAM chunk */
static int cx88_configure_fifos(cx88_channel *channel)
{
   DIAGNOSTIC(device_printf(channel->sc->dev, "Configuring SRAM FIFOs for channel %d\n", channel->descriptor->number))
   uint32_t fifo_space, used_fifo_space;
   // Set aside 1/32 of the chunk for the instruction queue */
   cx88_size_t instrq_size = channel->channel_size >> 5;
   cx88_size_t i;
   cx88_size_t min_num_fifos;
   channel->num_fifos = 0;
   instrq_size -= (instrq_size % 8);
   if (instrq_size < CX88_MIN_INSTRQ_SIZE)
   {
      instrq_size = CX88_MIN_INSTRQ_SIZE;
   }
   fifo_space = channel->channel_size - instrq_size;
   /* We need at least 2 FIFOs, each with space for at least one packet */
   if (fifo_space < (2 * channel->packet_size))
   {
      device_printf(channel->sc->dev, "SRAM allocation too small for channel %d\n", channel->descriptor->number);
      return ENOMEM;
   }

   /* Initialize the variable to track the amount of SRAM we're trying to use for FIFOs,
    * make sure it's a multiple of the packet size */
   used_fifo_space = fifo_space - (2 * CX88_CDT_ENTRY_SIZE);
   used_fifo_space -= (used_fifo_space % channel->packet_size);
   /* Loop until we've either calculated an optimum FIFOs size & number of FIFOs, or we
    * can't possibly use any less of the SRAM chunk */
   while ((used_fifo_space >= (2 * channel->packet_size)) && (channel->num_fifos == 0))
   {
      /* Step 1: Calculate the minimum number of FIFOs that we can possibly get away with.
       * It must be at least 2, and may be greater given maximum line size constraints */
      min_num_fifos = used_fifo_space / channel->max_line_size;
      if ((used_fifo_space % channel->max_line_size) != 0)
      {
         ++min_num_fifos;
      }
      if (min_num_fifos < 2)
      {
         min_num_fifos = 2;  
      }
      
      /* Step 2: Loop until we've found a way to divide the usable FIFO space */
      for (i = min_num_fifos; (i <= ((used_fifo_space / channel->packet_size)) && (i <= CX88_MAX_NUM_FIFOS)); ++i)
      {
         if ((used_fifo_space + (i * CX88_CDT_ENTRY_SIZE)) > fifo_space)
         {
            /* The required CDT space will only grow as i increases, so if we
             * can't fit them in now, just bail out of the loop */
            break;
         }
         cx88_size_t num_packets = used_fifo_space / channel->packet_size;
         /* The number of packets that can be stored in our FIFO space must
          * be divisible by the number of FIFOs, and the number of packets in each line
          * must not cross the specified boundary */
         if (((num_packets % i) == 0) && ((channel->line_boundary % (num_packets / i)) == 0))
         {
            cx88_size_t line_size = used_fifo_space / i;
            cx88_size_t fifo_size = line_size;
            /* The cx88 takes FIFO sizes in qwords, so make sure we align to an 8-byte boundary */
            if ((fifo_size % 8) != 0)
            {
               fifo_size += (8 - (fifo_size % 8));
            }
            /* Now make sure the aligned FIFOs and CDTs will fit in our SRAM chunk */
            if ((i * (CX88_CDT_ENTRY_SIZE + fifo_size)) <= fifo_space)
            {
               channel->fifo_size = fifo_size;
               channel->line_size = line_size;
               channel->num_fifos = i;
               break;
            }
         }
      }
      /* If we haven't yet found a suitable FIFO arrangement, subtract 
       * a packet and go back around */
      used_fifo_space -= channel->packet_size;
   }
   if (channel->num_fifos == 0)
   {
      device_printf(channel->sc->dev, "Unable to allocate SRAM for fifos for channel %d\n", channel->descriptor->number);
      return ENOMEM;
   }
   channel->cdt_offset = channel->channel_offset;
   channel->fifo_offset = channel->cdt_offset + (channel->num_fifos * CX88_CDT_ENTRY_SIZE);
   DIAGNOSTIC(
      device_printf(
         channel->sc->dev,
         "Channel %d:\n"
         "\tCDT offset = 0x%x\n"
         "\tNumber of FIFOs = %d\n"
         "\tFIFO offset = 0x%x\n"
         "\tFIFO size = 0x%x\n"
         "\tLine size = 0x%x\n",
         channel->descriptor->number,
         channel->cdt_offset,
         channel->num_fifos,
         channel->fifo_offset,
         channel->fifo_size,
         channel->line_size
         )
      )
   return 0;
}

/* bus_dma map load callback for RISC program buffers */
static void cx88_dmaprogmap_load(
   void *arg,
   bus_dma_segment_t *sgl,
   int sgl_size,
   int error)
{
   cx88_buffer *buffer = arg;
   /* We only allow physically contiguous RISC programs */
   buffer->dma_program_physaddr = (uint32_t) (sgl->ds_addr);
}

/* Allocates a DMA program buffer */
static int cx88_alloc_dmaprog_buffer(cx88_buffer *buffer)
{
   int error = 0;
   DIAGNOSTIC(printf("Allocating DMA program buffer\n"))
   cx88_release_dmaprogram(buffer);
   /* We only use physically contiguous buffers--shouldn't be more than a page or two
    * at most anyway.  Don't use bus_dmamem_alloc, as it always tries to allocate the
    * maximum buffer for the tag, which we may very well not need. */
   buffer->dma_program_buffer = contigmalloc(buffer->dma_program_size, M_DEVBUF, M_WAITOK,
      0, BUS_SPACE_MAXADDR_32BIT, CX88_DMA_ALIGNMENT, CX88_DMA_BOUNDARY);
   if (buffer->dma_program_buffer == NULL)
   {
      device_printf(buffer->channel->sc->dev, "Unable to allocate DMA program buffer\n");
      return ENOMEM;
   }
   error = bus_dmamap_create(buffer->channel->dma_program_tag, 0, &(buffer->dma_program_map));
   if (error)
   {
      device_printf(buffer->channel->sc->dev, "Unable to create DMA program buffer mapping\n");
      contigfree(buffer->dma_program_buffer, buffer->dma_program_size, M_DEVBUF);
      buffer->dma_program_buffer = NULL;
      return error;
   }
   error = bus_dmamap_load(
      buffer->channel->dma_program_tag,
      buffer->dma_program_map,
      buffer->dma_program_buffer,
      buffer->dma_program_size,
      cx88_dmaprogmap_load,
      buffer,
      BUS_DMA_WAITOK);
   if (error == EINPROGRESS)
   {
      /* If we have to wait on allocation, do so synchronously by waiting for the 
       * bus_dma lock callback to release the semaphore */
      sema_wait(&(buffer->channel->map_semaphore));
      error = 0;
   }
   if (error)
   {
      DIAGNOSTIC(printf("Unable to load DMA program map\n"))
      cx88_release_dmaprogram(buffer);
   }
   return error;
}

/* Allocate DMA and generate DMA programs for all buffers in the channel */
static int cx88_alloc_dmaprogram(cx88_channel *channel)
{
   int error = 0;
   uint8_t i;
   /* Whatever isn't used by the FIFOs goes to the instruction queue by default */
   channel->instrq_offset = channel->fifo_offset + (channel->fifo_size * channel->num_fifos);
   channel->instrq_size = channel->channel_offset + channel->channel_size - channel->instrq_offset;
   channel->line_offset = 0;
   for (i = 0; i < channel->num_buffers; ++i)
   {
      cx88_buffer *buffer = &(channel->buffers[i]);
      /* Make sure the data buffer has actually been allocated */
      if (buffer->data_buffer_type != CX88_DATABUF_NONE)
      {
         uint32_t old_program_size = buffer->dma_program_size;
         /* If the data buffer has been allocated but the DMA program hasn't
          * yet been generated, the size for the DMA program will need to be calculated.
          * The client-supplied generate_dma_program() callback needs to take this
          * state into account and only increase the DMA program counter without
          * actually writing the DMA program.  In reality, client callbacks shouldn't
          * need to worry about dma_program_state as they will build DMA programs through
          * the cx88_risc* functions, which correctly account for it. */
         if (buffer->dma_program_state == CX88_DMAPROG_SIZE)
         {
            cx88_init_dma_counters(buffer);
            if ((error = buffer->generate_dma_program(buffer, i)))
            {
               return error;
            }
            buffer->dma_program_size = buffer->dma_program_counter * sizeof (*(buffer->dma_program_buffer));
         }
         if (buffer->dma_program_size <= (channel->instrq_size - CX88_MIN_INSTRQ_SIZE))
         {
            /* The instruction queue is actually large enough to house this entire DMA program
             * with enough space left to maintain the minimum safe size, so go ahead and store
             * the program in SRAM */
            DIAGNOSTIC(device_printf(channel->sc->dev, "Allocating DMA program of size %d in SRAM\n", buffer->dma_program_size))
            cx88_release_dmaprogram(buffer);
            buffer->dma_program_physaddr = channel->instrq_offset;
            channel->instrq_size -= buffer->dma_program_size;
            channel->instrq_offset += buffer->dma_program_size;
         }
         else
         {
            /* We can't fit the DMA program in SRAM, so just allocate it in RAM */
            DIAGNOSTIC(device_printf(channel->sc->dev, "Allocating DMA program of size %d in RAM\n", buffer->dma_program_size))
            if ((buffer->dma_program_buffer == NULL) || (old_program_size != buffer->dma_program_size))
            {
               error = cx88_alloc_dmaprog_buffer(buffer);
               if (error) 
               {
                  device_printf(channel->sc->dev, "Unable to allocate DMA program buffer\n");
                  return error;
               }
            }
         }
      }
   }
   if (channel->instrq_size > CX88_MAX_INSTRQ_SIZE)
   {
      channel->instrq_size = CX88_MAX_INSTRQ_SIZE;
   }
   DIAGNOSTIC(
      device_printf(
         channel->sc->dev,
         "Channel %d:\n"
         "\tInstrq offset = 0x%x\n"
         "\tInstrq size = 0x%x\n",
         channel->descriptor->number,
         channel->instrq_offset,
         channel->instrq_size
         )
      )
   /* Now that we've sized and allocated the DMA programs, it's time
    * to actually generate them.  Change the state to CX88_DMAPROG_LOAD,
    * and invoke the client-supplied generator function again */
   for (i = 0; i < channel->num_buffers; ++i)
   {
      cx88_buffer *buffer = &(channel->buffers[i]);
      if (buffer->data_buffer_type != CX88_DATABUF_NONE)
      {
         buffer->dma_program_state = CX88_DMAPROG_LOAD;
         cx88_init_dma_counters(buffer);
         if ((error = buffer->generate_dma_program(buffer, i)))
         {
            return error;
         }
         cx88_prepare_for_transfer(buffer);
      }
   }
   return 0;
}

/* Commit the FIFO, instruction queue, and DMA program info to the cx88 registers.
 * After this, we're ready to transfer data */
int cx88_commit_channel(cx88_channel *channel)
{
   DIAGNOSTIC(device_printf(channel->sc->dev, "Committing channel %d configuration to registers\n", channel->descriptor->number))
   if (!channel->is_dirty)
   {
      return 0;
   }
   if ((channel->start_buffer >= channel->num_buffers) ||
      (channel->buffers[channel->start_buffer].data_buffer_type == CX88_DATABUF_NONE))
   {
      return EINVAL;
   }
   int error = cx88_alloc_dmaprogram(channel);
   if (error)
   {
      cx88_release_channel(channel);
      return error;
   }
   uint32_t cdt_size = (channel->num_fifos * CX88_CDT_ENTRY_SIZE) >> 3;
   uint32_t instrq_size = channel->instrq_size >> 2;
   uint32_t cmds_offset = CX88_CMDS(channel->descriptor->number);
   uint32_t reg_PTR2 = CX88_REG_PTR2(channel->descriptor->number);
   uint32_t reg_CNT1 = CX88_REG_CNT1(channel->descriptor->number);
   uint32_t reg_CNT2 = CX88_REG_CNT2(channel->descriptor->number);
   cx88_size_t i;
   if (channel->buffers[channel->start_buffer].dma_program_buffer == NULL)
   {
      /* The cx88 stores the flag indicating an SRAM resident DMA program
       * in the same word as the instruction queue size */
      instrq_size |= (1 << 31);
   }
   for (i = 0; i < channel->num_fifos; ++i)
   {
      /* Write the FIFO base addresses to the CDTs */
      cx88_write_reg(channel->sc, 
         (channel->cdt_offset + (i * CX88_CDT_ENTRY_SIZE)),
         (channel->fifo_offset + (i * channel->fifo_size))); 
      DIAGNOSTIC(
         device_printf(
            channel->sc->dev,
            "Writing FIFO base address 0x%x to SRAM offset 0x%x\n",
            (channel->fifo_offset + (i * channel->fifo_size)),
            (channel->cdt_offset + (i * CX88_CDT_ENTRY_SIZE))
            )
         )
   }
   /* Generate the CMDS */
   cx88_write_reg(channel->sc, cmds_offset, channel->buffers[channel->start_buffer].dma_program_physaddr);
   cx88_write_reg(channel->sc, cmds_offset + 0x4, channel->cdt_offset);
   cx88_write_reg(channel->sc, cmds_offset + 0x8, cdt_size);
   cx88_write_reg(channel->sc, cmds_offset + 0xC, channel->instrq_offset);
   cx88_write_reg(channel->sc, cmds_offset + 0x10, instrq_size);
   /* Write CDT and FIFO info */
   cx88_write_reg(channel->sc, reg_PTR2, channel->cdt_offset);
   cx88_write_reg(channel->sc, reg_CNT1, ((channel->fifo_size >> 3) - 1));
   cx88_write_reg(channel->sc, reg_CNT2, cdt_size);
   channel->is_dirty = FALSE;
   return 0;
}

/* Configure channel properties */
int cx88_configure_channel(
   cx88_channel *channel,
   cx88_size_t packet_size,
   cx88_size_t max_line_size,
   uint32_t line_boundary)
{
   DIAGNOSTIC(device_printf(channel->sc->dev, "Configuring channel %d\n", channel->descriptor->number))
   int error = 0;
   /* Allocate the channel if it doesn't already have an SRAM chunk */
   if ((channel->sram_region == NULL) && (error = cx88_alloc_channel(channel)))
   {
      return error;
   }
   if ((channel->packet_size != packet_size) || (channel->max_line_size != max_line_size)
      || (channel->line_boundary != line_boundary))
   {
      /* If the packet size, the maximum line size, or the line boundary are changing,
      * we must re-calculate the FIFO arrangement */
      cx88_release_buffers(channel);
      channel->packet_size = packet_size;
      channel->max_line_size = max_line_size;
      channel->line_boundary = line_boundary;
      error = cx88_configure_fifos(channel);
      channel->is_dirty = TRUE;
   }
   return error;
}

/* Request a given number of buffers for a channel */
int cx88_request_buffers(
   cx88_channel *channel, 
   uint8_t num_buffers,
   uint8_t start_buffer)
{
   uint8_t i;
   uint8_t prev_buffers = channel->num_buffers;
   cx88_buffer *requested_bufs;
   if (prev_buffers != num_buffers)
   {
      requested_bufs = realloc(channel->buffers, sizeof(*(channel->buffers)) * num_buffers, M_DEVBUF, M_WAITOK);
      if (requested_bufs == NULL)
      {
         return ENOMEM;
      }
      channel->is_dirty = TRUE;
      if (prev_buffers > num_buffers)
      {
         /* Release resources for any existing buffers
          * beyond the number we're requesting */
         for (i = num_buffers; i < prev_buffers; ++i)
         {
            cx88_release_buffer(&(channel->buffers[i]));
         }
      }
      channel->buffers = requested_bufs;
      channel->num_buffers = num_buffers;
      /* Initialize any buffers that have been added to the buffer array */
      for (i = prev_buffers; i < num_buffers; ++i)
      {
         cx88_buffer * buffer = &(channel->buffers[i]);
         buffer->channel = channel;
         buffer->data_buffer_size = 0;
         buffer->data_buffer_type = CX88_DATABUF_NONE;
         buffer->data_buffer = NULL;
         buffer->dma_program_size = 0;
         buffer->dma_program_buffer = NULL;
         buffer->generate_dma_program = NULL;
         buffer->sgl = NULL;
         buffer->next_buffer = NULL;
         cx88_init_dma_counters(buffer);
      }
   }
   if (start_buffer >= channel->num_buffers)
   {
      return EINVAL;
   }
   else if (start_buffer != channel->start_buffer)
   {
      channel->start_buffer = start_buffer;
      channel->is_dirty = TRUE;
   }
   return 0;
}

/* bus_dma locking callback for M_WAITOK mappings */
static void cx88_dmamap_lock(void *arg, bus_dma_lock_op_t op)
{
   /* For both data and RISC buffers, we wait synchronously
    * on a semaphore */
   if (op == BUS_DMA_UNLOCK)
   {
      sema_post((struct sema*)arg);
   }
}

/* bus_dma map load callback for data buffers */
static void cx88_dmamap_load(
   void *arg,
   bus_dma_segment_t *sgl,
   int sgl_size,
   int error)
{
   cx88_buffer *buffer = arg;
   DIAGNOSTIC(printf("Allocating SGL of size %d\n", sgl_size))
   if (error)
   {
      return;
   }
   /* Store the scatter/gather list in the buffer */
   if ((buffer->sgl = malloc(sizeof(*sgl) * sgl_size, M_DEVBUF, M_WAITOK)) == NULL)
   {
      printf("CX88: Unable to allocate scatter/gather list\n");
      return;
   }
   memcpy(buffer->sgl, sgl, sizeof(*sgl) * sgl_size);
   buffer->sgl_size = sgl_size;
   buffer->dma_program_state = CX88_DMAPROG_SIZE;
}

/* bus_dma map load callback for user-allocated buffers */
static void cx88_dmamap_load_uio(
   void *arg,
   bus_dma_segment_t *sgl,
   int sgl_size,
   bus_size_t mapsize,
   int error)
{
   cx88_dmamap_load(arg, sgl, sgl_size, error);
}

/* Configure a buffer structure and allocate its data buffer */
int cx88_configure_buffer(
   cx88_buffer *buffer,
   void *userbuf,
   struct thread * td,
   uint32_t num_packets,
   cx88_buffer *next_buffer,
   cx88_risc_generator dma_program_generator)
{
   DIAGNOSTIC(device_printf(buffer->channel->sc->dev, "Configuring DMA data buffer\n"))
   bus_size_t buffer_size = buffer->channel->packet_size * num_packets;
   if ((buffer->generate_dma_program != dma_program_generator) ||
      (next_buffer != buffer->next_buffer))
   {
      buffer->generate_dma_program = dma_program_generator;
      buffer->next_buffer = next_buffer;
      buffer->channel->is_dirty = TRUE;
   }
   if ((buffer->data_buffer_size == buffer_size) &&
      (buffer->data_buffer_type != CX88_DATABUF_USER) && (userbuf == NULL))
   {
      return 0;
   }
   buffer->channel->is_dirty = TRUE;
   cx88_release_buffer(buffer);
   buffer->data_buffer_size = buffer_size;
   int error = bus_dmamap_create(buffer->channel->data_buffer_tag, 0, &(buffer->data_buffer_map));
   if (error)
   {
      device_printf(buffer->channel->sc->dev, "Unable to create DMA map for buffer\n");
      return error;
   }
   if (userbuf == NULL) /* Allocate buffer in kernel mode */
   {
      /* As with RISC buffers, don't use bus_dmamem_alloc as it will allocate the maximum allowable
       * buffer size, which we probably don't need.  And it uses contigmalloc() underneath.  Here we
       * use malloc() to ensure greater flexibility in allocating larger data buffers, at the possible
       * expense of bounce buffering on non-IOMMU machines with >4GB of RAM */
      if ((buffer->data_buffer = malloc(buffer->data_buffer_size, M_DEVBUF, M_WAITOK | M_ZERO)) == NULL)
      {
         device_printf(buffer->channel->sc->dev, "Unable to allocate DMA buffer for buffer\n");
         bus_dmamap_destroy(buffer->channel->data_buffer_tag, buffer->data_buffer_map);
         return ENOMEM;
      }
      buffer->data_buffer_type = CX88_DATABUF_MALLOC;
      error = bus_dmamap_load(
         buffer->channel->data_buffer_tag,
         buffer->data_buffer_map,
         buffer->data_buffer,
         buffer->data_buffer_size,
         cx88_dmamap_load,
         buffer,
         BUS_DMA_WAITOK);
   }
   else /* Buffer was allocated in user mode */
   {
      error = vslock(userbuf, buffer->data_buffer_size);
      if (error)
      {
         device_printf(buffer->channel->sc->dev, "Unable to wire buffer\n");
         bus_dmamap_destroy(buffer->channel->data_buffer_tag, buffer->data_buffer_map);
         return error;
      }
      buffer->data_buffer = userbuf;
      buffer->data_buffer_type = CX88_DATABUF_USER;
      struct iovec uvec;
      struct uio ubuf;
      uvec.iov_base = buffer->data_buffer;
      uvec.iov_len = buffer->data_buffer_size;
      ubuf.uio_iov = &uvec;
      ubuf.uio_iovcnt = 1;
      ubuf.uio_segflg = UIO_USERSPACE;
      ubuf.uio_td = td;
      ubuf.uio_resid = buffer->data_buffer_size;
      ubuf.uio_offset = 0;
      ubuf.uio_rw = ((buffer->channel->descriptor->direction == CX88_DOWNSTREAM) ? UIO_WRITE : UIO_READ);
      error = bus_dmamap_load_uio(
         buffer->channel->data_buffer_tag,
         buffer->data_buffer_map,
         &ubuf,
         cx88_dmamap_load_uio,
         buffer,
         BUS_DMA_WAITOK);
   }
   if (error == EINPROGRESS)
   {
      sema_wait(&(buffer->channel->map_semaphore));
      error = 0;
   }
   if (!error && (buffer->sgl == NULL))
   {
      error = ENOMEM;
   }
   if (error)
   {
      cx88_release_buffer(buffer);
      device_printf(buffer->channel->sc->dev, "Unable to load DMA map for buffer\n");
   }
   return error;
}

/* Synchronize RISC and data buffers before DMA */
int cx88_prepare_for_transfer(cx88_buffer *buffer)
{
   if (buffer->channel->is_dirty)
   {
      return EINVAL;
   }
   if (buffer->dma_program_buffer != NULL)
   {
      bus_dmamap_sync(buffer->channel->dma_program_tag, buffer->dma_program_map, BUS_DMASYNC_PREWRITE);
   }
   bus_dmamap_sync(
      buffer->channel->data_buffer_tag,
      buffer->data_buffer_map,
      ((buffer->channel->descriptor->direction == CX88_DOWNSTREAM) ? 
         BUS_DMASYNC_PREREAD : BUS_DMASYNC_PREWRITE));
   return 0;
}

/* Synchronize RISC and data buffers after DMA */
void cx88_mark_transfer_complete(cx88_buffer *buffer)
{
   if (buffer->dma_program_buffer != NULL)
   {
      bus_dmamap_sync(buffer->channel->dma_program_tag, buffer->dma_program_map, BUS_DMASYNC_POSTWRITE);
   }
   bus_dmamap_sync(
      buffer->channel->data_buffer_tag,
      buffer->data_buffer_map,
      ((buffer->channel->descriptor->direction == CX88_DOWNSTREAM) ?
         BUS_DMASYNC_POSTREAD : BUS_DMASYNC_POSTWRITE));
}

/* Probe routine */
int cx88_common_probe(device_t dev, const cx88_client_driver * client) 
{
   uint16_t pid = pci_get_device(dev);
   uint16_t vid = pci_get_vendor(dev);
   /* Match against Conexant's vendor ID and a product ID generated
    * from the client's function index */
   if ((vid == CX88_VID) && 
      (pid == (client->function + CX88_BASE_PID)) &&
      (client->function < CX88_NUM_FUNCTIONS))
   {
      device_set_desc(dev, client->description);
      return BUS_PROBE_DEFAULT;
   }
   return ENXIO;
}

/* Free a common softc structure.  Used both for device detachment
 * and error cases during device attachment, so must check
 * each softc item before freeing it */
static int cx88_release_softc(cx88_common_softc *sc)
{
   int error = 0;
   if (sc->irq_cookie != NULL)
   {
      if ((error = bus_teardown_intr(sc->dev, sc->irq_res, sc->irq_cookie)) != 0)
      {
         return error;
      }
      sc->irq_cookie = NULL;
   }
   if (sc->is_configured && (sc->client->teardown != NULL))
   {
      if ((error = sc->client->teardown(sc)) != 0)
      {
         return error;
      }
      sc->is_configured = FALSE;
   }
   if (sc->irq_res != NULL)
   {
      bus_release_resource(sc->dev, SYS_RES_IRQ, sc->irq_rid, sc->irq_res);
      sc->irq_res = NULL;
   }
   if (sc->mem_res != NULL)
   {
      bus_release_resource(sc->dev, SYS_RES_MEMORY, sc->mem_rid, sc->mem_res);
      sc->mem_res = NULL;
   }
   if (sc->board_data != NULL)
   {
      cx88_lock_board(sc);
      sc->board_data->present_channels &= (~(sc->client->channel_mask));
      cx88_unlock_board(sc);
      if ((error = cx88_remove_board_data(sc, sc->board_data, cx88_destroy_board_data)) != 0)
      {
         return error;
      }
      sc->board_data = NULL;
   }
   /* Softc is externally allocated (see below) */
   device_set_softc(sc->dev, NULL);
   free(sc, M_DEVBUF);
   return error;
}

/* Attach routine */
int cx88_common_attach(device_t dev, const cx88_client_driver *client) 
{
   /* Externally allocate softc so that we can set it to NULL during detachment
    * without freeing it--since our DEVMETHODs will still be valid, DEVMETHODs
    * that may be called from other threads can use cx88_get_softc() to
    * determine that the device is no longer in a valid state */
   cx88_common_softc *sc = malloc(client->softc_size, M_DEVBUF, M_NOWAIT | M_ZERO);
   uint32_t cmdr;
   uint8_t i;
   int error = 0;
   if (sc == NULL)
   {
      return ENOMEM;
   }
   device_set_softc(dev, sc);
   sc->board_data = NULL;
   sc->mem_res = NULL;
   sc->irq_res = NULL;
   sc->irq_cookie = NULL;
   sc->client = client;
   sc->is_configured = FALSE;
   sc->dev = dev;
   sc->num_siblings = 0;
   /* Generate the array of sibling devices */
   for (i = 0; i < CX88_NUM_FUNCTIONS; ++i)
   {
      device_t sibling;
      /* Skip our own function */
      if (i == client->function)
      {
         continue;
      }
      sibling = pci_find_bsf(pci_get_bus(dev), pci_get_slot(dev), i);
      if (sibling != NULL)
      {
         DIAGNOSTIC(device_printf(dev, "Found sibling at function %d\n", i))
         sc->siblings[(sc->num_siblings)++] = sibling;
      }
   }
   cmdr = pci_read_config(dev, PCIR_COMMAND, 2);
   /* Enable memory window, bus mastering, and write caching */
   cmdr |= (PCIM_CMD_MEMEN | PCIM_CMD_BUSMASTEREN | PCIM_CMD_MWRICEN);
   pci_write_config(dev, PCIR_COMMAND, cmdr, 2);
   sc->mem_rid = PCIR_BAR(0);
   sc->mem_res = bus_alloc_resource_any(dev, SYS_RES_MEMORY, &sc->mem_rid, RF_ACTIVE);
   if (sc->mem_res == NULL)
   {
      device_printf(dev, "Unable to map memory\n");
      cx88_release_softc(sc);
      return ENXIO;
   }
   sc->mem_tag = rman_get_bustag(sc->mem_res);
   sc->mem_handle = rman_get_bushandle(sc->mem_res);
   if (pci_read_config(dev, PCIR_LATTIMER, 1) == 0)
   {
      pci_write_config(dev, PCIR_LATTIMER, CX88_DEFAULT_LATENCY, 1);
   }
   sc->irq_rid = 0;
   sc->irq_res = bus_alloc_resource_any(dev, SYS_RES_IRQ, &sc->irq_rid, RF_SHAREABLE | RF_ACTIVE);
   if (sc->irq_res == NULL)
   {
      device_printf(dev, "Unable to allocate interrupt resource\n");
      cx88_release_softc(sc);
      return ENXIO;
   }
   /* Load the common board data structure */
   if ((error = cx88_add_board_data(sc, (void **)(&(sc->board_data)),
      sizeof(*(sc->board_data)), cx88_get_sibling_board_data, cx88_init_board_data)))
   {
      cx88_release_softc(sc);
      return error;
   }
   /* Modify the channel bitmask in the board data structure */
   cx88_lock_board(sc);
   if (sc->board_data->present_channels & sc->client->channel_mask)
   {
      device_printf(dev, "Channel conflict detected\n");
      error = EINVAL;
   }
   else
   {
      sc->board_data->present_channels |= sc->client->channel_mask;  
   }
   cx88_unlock_board(sc);
   if (error)
   {
      cx88_release_softc(sc);
      return error;
   }
   /* Call client-specific softc setup routine */
   if ((client->setup != NULL) && ((error = client->setup(sc))))
   {
      cx88_release_softc(sc);
      return error;
   }
   sc->is_configured = TRUE;
   if ((error = bus_setup_intr(dev, sc->irq_res, INTR_TYPE_AV | INTR_MPSAFE,
      client->interrupt_handler, sc, &(sc->irq_cookie))))
   {
      device_printf(dev, "Unable to establish interrupt handler\n");
      cx88_release_softc(sc);
      return error;
   }
   return 0;
}

/* Detach routine */
int cx88_common_detach(device_t dev)
{
   int error = 0, i;
   cx88_common_softc *sc = device_get_softc(dev);
   cx88_lock_board(sc);
   if (device_get_state(sc->dev) == DS_BUSY)
   {
      error = EBUSY;
   }
   /* Prevent detach if even a sibling is busy--that
    * sibling may otherwise try to call our DEVMETHODs 
    * while we are tearing down */
   for (i = 0; ((error == 0) && (i < sc->num_siblings)); ++i)
   {
      if (device_get_state(sc->siblings[i]) == DS_BUSY)
      {
         error = EBUSY;
      }
   }
   if (!error)
   {
      device_set_softc(dev, NULL);
   }
   cx88_unlock_board(sc);
   if (!error)
   {
      if ((error = cx88_release_softc(sc)))
      {
         /* Reset the softc so that we can try releasing it later */
         device_set_softc(dev, sc);  
      }
   }
   if (!error)
   {
      bus_generic_detach(dev);
   }
   return error;
}

/* Initializes a newly-allocated cx88_board_data structure.
 * Called when the current client is the first driver to reference
 * the board data */
static int cx88_init_board_data(void *arg)
{
   int error = 0;
   cx88_board_data *data = arg;
   data->sram.rm_start = 0;
   data->sram.rm_end = CX88_USABLE_SRAM - 1;
   data->sram.rm_type = RMAN_ARRAY;
   data->sram.rm_descr = "CX88 SRAM";
   error = rman_init(&(data->sram));
   if (!error)
   {
      if ((error = rman_manage_region(&(data->sram), 0, CX88_USABLE_SRAM - 1)))
      {
         rman_fini(&(data->sram));  
      }
   }
   mtx_init(&(data->lock), "CX88 board lock", NULL, MTX_DEF | MTX_RECURSE);
   data->present_channels = 0;
   return error;
}

/* Cleans up the board data structure when the current client is the
 * last to reference it and it is about to be freed */
static int cx88_destroy_board_data(void *arg)
{
   cx88_board_data *data = arg;
   mtx_destroy(&(data->lock));
   rman_fini(&(data->sram));
   return 0;
}

/* Accessor function for board data on siblings */
static void *cx88_get_sibling_board_data(device_t dev)
{
   return CX88_GET_BOARD_DATA(dev);
}

/* Generic function for adding data to be shared by multiple
 * drivers for the same board */
int cx88_add_board_data(
   cx88_common_softc *sc,
   void **data,
   size_t data_size,
   get_board_data_func accessor,
   cx88_setup_func setup
   )
{
   /* Ensure we can add or update the board data structure atomically
    * w/ respect to siblings */
   mtx_lock(&interdev_lock);
   *data = NULL;
   int error = 0, i;
   /* First check if a siblings has allocated the structure */
   for (i = 0; i < sc->num_siblings; ++i)
   {
      if ((*data = accessor(sc->siblings[i])) != NULL)
      {
         DIAGNOSTIC(device_printf(sc->dev, "Found data on sibling board\n"))
         break;
      }
   }
   if (*data == NULL)
   {
      /* No sibling has allocated the structure, so we need to */
      size_t *base_addr;
      if ((base_addr = malloc(data_size + sizeof(size_t), M_DEVBUF, M_NOWAIT | M_ZERO)) == NULL)
      {
         device_printf(sc->dev, "Unable to allocate board data\n");
         mtx_unlock(&interdev_lock);
         return ENOMEM;
      }
      DIAGNOSTIC(device_printf(sc->dev, "Allocated board data\n"))
      /* Place the refcount for the structure in the word immediately preceding the 
       * the address that will be returned to the client */
      *base_addr = 1;
      *data = base_addr + 1;
      /* If the client supplied an initialization routine, call it now */
      if (setup != NULL && (error = setup(*data)))
      {
         free(base_addr, M_DEVBUF);
         *data = NULL;
      }
   }
   else
   {
      /* A sibling has already allocated the structure, so just increment its
       * refcount */
      size_t *base_addr = ((size_t *) (*data)) - 1;
      ++(*base_addr);
   }
   mtx_unlock(&interdev_lock);
   return error;
}

/* Function for removing a reference to a shared data structure */
int cx88_remove_board_data(
   cx88_common_softc *sc,
   void *data,
   cx88_teardown_func teardown
   )
{
   int error = 0;
   mtx_lock(&interdev_lock);
   if (data != NULL)
   {
      size_t *base_addr = ((size_t *)data) - 1;
      if (*base_addr == 1)
      {
         /* The refcount is 1, so we're the last driver to use the structure.
          * Go ahead and free it */
         DIAGNOSTIC(device_printf(sc->dev, "Freeing board data\n"))
         if (teardown != NULL)
         {
            /* If the client supplied a cleanup routine, call it now */
            error = teardown(data);
         }
         if (!error)
         {
            free(base_addr, M_DEVBUF);
         }
      }
      else
      {
         /* We're not the last to use it, so just decrement the refcount */
         --(*base_addr);
      }
   }
   mtx_unlock(&interdev_lock);
   return error;
}

/* Mark a device busy if it is not busy already.
 * Do this atomically with respect to siblings */
int cx88_busy(cx88_common_softc *sc)
{
   int error = 0;
   cx88_lock_board(sc);
   if (device_get_state(sc->dev) == DS_BUSY)
   {
      error = EBUSY;
   }
   else
   {
      device_busy(sc->dev);  
   }
   cx88_unlock_board(sc);
   return error;
}

/* Mark a device as no longer busy, atomically with
 * respect to siblings */
void cx88_unbusy(cx88_common_softc *sc)
{
   cx88_lock_board(sc);   
   device_unbusy(sc->dev);
   cx88_unlock_board(sc);
}

/* Retrieve a driver's softc, only if it has been
 * allocated and initialized */
void *cx88_get_softc(device_t dev)
{
   cx88_common_softc *sc = device_get_softc(dev);
   if ((sc == NULL) || !(sc->is_configured))
   {
      return NULL;
   }
   return sc;
}

/* DEVMETHOD for client drivers to implement so that siblings
 * can access their common board data structure */
cx88_board_data *cx88_common_get_board_data(device_t dev)
{
   cx88_common_softc *sc = device_get_softc(dev);
   if (sc != NULL)
   {
      return sc->board_data;
   }
   return NULL;
}

/* Generic mmap function that clients can call from their d_mmap routines */
int cx88_common_mmap(
   cx88_channel *channel,
   vm_offset_t offset,
   vm_paddr_t *paddr,
   int nprot)
{
   int i;
   bus_size_t phys_offset = 0;
   int error = EINVAL;
   if ((channel->descriptor->direction == CX88_DOWNSTREAM) && (nprot & PROT_WRITE))
   {
      return error;
   }
   for (i = 0; i < channel->num_buffers; ++i)
   {
      cx88_buffer *buffer = &(channel->buffers[i]);
      if ((buffer->data_buffer_type == CX88_DATABUF_NONE) ||
         (buffer->data_buffer_type == CX88_DATABUF_USER))
      {
         /* Only map kernel-allocated buffers */
         continue;
      }
      if ((phys_offset + buffer->data_buffer_size) > offset)
      {
         //Use vtophys instead of the stored sgl--we want the CPU-visible physical address
         //of the selected offset in the original buffer, while the sgl stores device-visible
         //physical addresses (they may differ on machines with IOMMUs or in cases where 
         //bounce buffers are used).
         *paddr = vtophys (((char*) (buffer->data_buffer)) + (offset - phys_offset));
         DIAGNOSTIC(device_printf(channel->sc->dev, "MMAP: Using buffer %d at offset 0x%lx\n", i, (offset - phys_offset)))
         return 0;
      }
      else
      {
         phys_offset += round_page(buffer->data_buffer_size);
      }
   }
   return error;
}

/* Write GPIO pins */
void cx88_write_gpio(cx88_common_softc *sc, uint32_t mask, uint32_t value)
{
   cx88_lock_board(sc);
   static uint32_t regs[] = {CX88_REG_GPIO0, CX88_REG_GPIO1, CX88_REG_GPIO2, CX88_REG_GPIO3};
   int i;
   uint8_t shift;
   uint32_t write_data = mask & value;
   uint32_t write_mask;
   for (i = 0; i < (sizeof(regs) / sizeof(uint32_t)); ++i)
   {
      shift = i * 8;
      write_mask = (mask >> shift) & 0xFF;
      if (write_mask != 0)
      {
         uint32_t field = 0;
         /* Enable writing for only the bits we want to write */
         field |= cx88_write_field(CX88_GPIO_BIT_WRITE, write_mask);
         field |= cx88_write_field(CX88_GPIO_OUT_ENABLE, write_mask);
         field |= cx88_write_field(CX88_GPIO_IO, (write_data >> shift));
         cx88_write_reg(sc, regs[i], field);
      }
   }
   cx88_unlock_board(sc);
}

/* Read GPIO pins */
uint32_t cx88_read_gpio(cx88_common_softc *sc, uint32_t mask)
{
   uint32_t value = 0;
   cx88_lock_board(sc);
   static uint32_t regs[] = {CX88_REG_GPIO0, CX88_REG_GPIO1, CX88_REG_GPIO2, CX88_REG_GPIO3};
   int i;
   uint8_t shift;
   uint32_t read_mask;
   for (i = 0; i < (sizeof(regs) / sizeof(uint32_t)); ++i)
   {
      shift = i * 8;
      read_mask = (mask >> shift) & 0xFF;
      if (read_mask != 0)
      {
         value |= (cx88_read_field(cx88_read_reg(sc, regs[i]), CX88_GPIO_IO) & read_mask);
      }
   }
   cx88_unlock_board(sc);
   return value;
}
