/*-
 * Copyright 2007 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
 
 
 /* File: cx88_common.h
  * Purpose: Header file for core cx88 channel, buffer, and driver management
  * functions
  */

#ifndef __CX88_COMMON_H__
#define __CX88_COMMON_H__

#include <sys/types.h>
#include <sys/param.h>
#include <sys/bus.h>
#include <machine/bus.h>
#include <sys/lock.h>
#include <sys/mutex.h>
#include <sys/proc.h>
#include <sys/sema.h>
#include <sys/rman.h>
#include <dev/pci/pcivar.h>

#ifdef _DIAGNOSTIC
#define DIAGNOSTIC(stmt) stmt;
#else
#define DIAGNOSTIC(stmt)
#endif

#define CX88_MINVER 1
#define CX88_MODVER 1
#define CX88_MAXVER 1

typedef uint16_t cx88_size_t; /* cx88 SRAM size */

/* cx88 PCI functions */
typedef enum
{
   CX88_FUNC_VIDEO = 0,
   CX88_FUNC_AUDIO,
   CX88_FUNC_MPEGTS,
   CX88_FUNC_VIP,
   CX88_FUNC_GPHOST,
   CX88_NUM_FUNCTIONS
} cx88_function_t;

#define CX88_MAX_SIBLINGS (CX88_NUM_FUNCTIONS - 1)

typedef int (*cx88_setup_func) (void *);
typedef int (*cx88_teardown_func) (void *);
typedef driver_intr_t (*cx88_intr_handler);

/* Defines vital statistics of a client driver */
typedef struct
{
   size_t softc_size; /* total size of the client's softc */
   cx88_function_t function; /* PCI function */
   uint16_t channel_mask; /* bitmask of cx88 DMA channels used by client */
   cx88_setup_func setup; /* initialization callback for client's softc */
   cx88_teardown_func teardown; /* cleanup callback for client's softc */
   cx88_intr_handler interrupt_handler; /* client's interrupt handler */
   const char * description; /* description of client */
} cx88_client_driver;

/* Data structure to be shared by all cx88 client drivers for
 * a particular card.  Allocated by first client driver to attach
 * and freed by last client to detach */
typedef struct
{
   struct mtx lock;
   struct rman sram;
   uint16_t present_channels;
} cx88_board_data;

/* Common softc structure. This MUST be the first member of each client's softc */
typedef struct
{
   device_t                  dev;
   device_t                  siblings[CX88_MAX_SIBLINGS];
   int                       num_siblings;
   bus_space_tag_t           mem_tag; 
   bus_space_handle_t        mem_handle;
   struct resource          *mem_res;
   int                       mem_rid;
   struct resource          *irq_res;
   int                       irq_rid;
   void                     *irq_cookie;
   const cx88_client_driver *client;
   cx88_board_data          *board_data;
   boolean_t                 is_configured;
} cx88_common_softc;

/* cx88 DMA channel numbers.  The cx88 datasheet numbers these
 * in the same order, but starting at 21 instead of 0 */
typedef enum
{
   CX88_CH_VIDEO_PACKED = 0,
   CX88_CH_VIDEO_U,
   CX88_CH_VIDEO_V,
   CX88_CH_VIDEO_VBI,
   CX88_CH_AUDIO_IN,
   CX88_CH_AUDIO_OUT,
   CX88_CH_UNUSED,
   CX88_CH_MPEG_TS,
   CX88_CH_VIP_IN,
   CX88_CH_VIP_OUT,
   CX88_CH_GPHOST_IN,
   CX88_CH_GPHOST_OUT
} cx88_channel_t;

/* Generates a bitmask value (uint16_t) for a channel number */
#define CX88_CHANNEL_MASK(channel) (1 << channel)

/* Vital data about a DMA channel */
typedef struct
{
   cx88_channel_t             number;
   enum
   {
      CX88_UPSTREAM,
      CX88_DOWNSTREAM
   }                          direction;
   cx88_size_t                   min_size;
} cx88_channel_desc;

typedef struct _cx88_buffer cx88_buffer;

typedef struct _cx88_channel cx88_channel;

/* Client drivers should provide functions for generating the 
 * DMA RISC program */
typedef int (*cx88_risc_generator) (cx88_buffer*, uint8_t);

/* Describes a data buffer and its DMA program. Each channel
 * may contain multiple buffers */
struct _cx88_buffer
{
   cx88_channel              *channel;
   enum
   {
      CX88_DMAPROG_SIZE,
      CX88_DMAPROG_LOAD
   }                          dma_program_state;
   uint32_t                  *dma_program_buffer;
   bus_dmamap_t               dma_program_map;
   uint32_t                   dma_program_physaddr;
   uint32_t                   dma_program_size;
   uint32_t                   dma_program_counter;
   uint32_t                   dma_byte_counter;
   enum
   {
      CX88_DATABUF_NONE,
      CX88_DATABUF_MALLOC,
      CX88_DATABUF_USER
   }                          data_buffer_type;
   void                      *data_buffer;
   bus_dmamap_t               data_buffer_map;
   bus_dma_segment_t         *sgl;
   int                        sgl_size;
   bus_dma_segment_t         *sgl_index;
   bus_size_t                 sgl_offset;
   bus_size_t                 data_buffer_size;
   cx88_risc_generator        generate_dma_program;
   cx88_buffer               *next_buffer;
};

/* Describes a DMA channel, its SRAM layout, and data buffers.
 * Client drivers are responsible for storing as many of these
 * structures as necessary in their softc structures.
 * All SRAM offsets are based on the beginning of the cx88 memory
 * space, not the beginning of the SRAM in that memory space */
struct _cx88_channel
{
   const cx88_channel_desc   *descriptor;
   cx88_common_softc         *sc;
   struct resource           *sram_region;
   uint32_t                   channel_offset;
   uint32_t                   instrq_offset;
   uint32_t                   cdt_offset;
   uint32_t                   fifo_offset;
   cx88_size_t                packet_size;
   cx88_size_t                line_size;
   cx88_size_t                line_offset;
   cx88_size_t                channel_size;
   cx88_size_t                instrq_size;
   cx88_size_t                fifo_size;
   cx88_size_t                num_fifos;
   cx88_buffer               *buffers;
   uint8_t                    num_buffers;
   uint8_t                    start_buffer;
   boolean_t                  is_dirty;
   bus_dma_tag_t              dma_program_tag;
   bus_dma_tag_t              data_buffer_tag;
   struct sema                map_semaphore;
   uint32_t                   max_buffer_size;
   uint32_t                   line_boundary; /* Number of packets, not bytes */
   cx88_size_t                max_line_size;
};

/* Probe function to be called by clients' probe DEVMETHODs.
 * Returns 0 on success, appropriate error otherwise */
int cx88_common_probe(device_t dev, const cx88_client_driver *client);

/* Attach function to be called by clients' attach DEVMETHODs 
 * Returns 0 on success, appropriate error otherwise */
int cx88_common_attach(device_t dev, const cx88_client_driver *client);

/* Detach function to be called by clients' detach DEVMETHODs
 * Returns 0 on success, appropriate error otherwise */
int cx88_common_detach(device_t dev);

/* mmap function to be called by clients' d_mmap
 * Returns 0 on success, appropriate error otherwise */
int cx88_common_mmap(
   cx88_channel *channel, /* Channel whose buffers should be mapped */
   vm_offset_t offset,
   vm_paddr_t *paddr,
   int nprot);

/* Clients should call this to configure a buffer in a channel's 
 * buffer array.  Returns 0 on success, appropriate error otherwise. */
int cx88_configure_buffer(
   cx88_buffer *buffer,
   void *userbuf, /* User-allocated buffer.  If NULL, buffer will be allocated in kernel*/
   struct thread *td, /* Must be non-NULL if userbuf is non-NULL */
   uint32_t num_packets, /* Number of packets the buffer must store */
   cx88_buffer *next_buffer, /* Next buffer in RISC program sequence */
   cx88_risc_generator dma_program_generator); /* Generates this buffer's RISC program */

/* Frees all resources associated with a buffer */
void cx88_release_buffer(cx88_buffer *buffer);

/* Allocates an initializes buffers for a channel.  Each buffer should
 * then be configured with cx88_configure_buffer(). Returns 0 on success, appropriate
 * error otherwise. */
int cx88_request_buffers(
   cx88_channel *channel,
   uint8_t num_buffers,
   uint8_t start_buffer); /* Index of buffer whose RISC program will run first when DMA is started */
   
/* Calls cx88_release_buffer() on all buffers associated with a channel */
void cx88_release_buffers(cx88_channel *channel);

/* Configures a DMA channel. Returns 0 on success, appropriate error otherwise. */
int cx88_configure_channel(
   cx88_channel *channel,
   cx88_size_t packet_size, /* Packet size (or line size) for channel */
   cx88_size_t max_line_size, /* Maximum allowable line size (in bytes).  Must be equal to packet_size for line-based channels */
   uint32_t line_boundary); /* Lines in a buffer must not cross multiple of this no. of packets (0 -> no boundary) */
   
/* Frees all resources associated with a chanel, including buffers */
void cx88_release_channel(cx88_channel *channel);

/* Once a channel and its buffers have been configured, generate the DMA programs and
 * write the channel data to the cx88 registers and CMDS. Returns 0 on success, appropriate
 * error otherwise. */
int cx88_commit_channel(cx88_channel *channel);

/* Initialize a channel structure in client's softc.  Most likely called
 * from the client's softc setup function during driver attachment.  Returns
 * 0 on success, appropriate error otherwise. */
int cx88_init_channel(
   cx88_channel *channel,
   cx88_common_softc *sc,
   cx88_channel_t channel_number,
   uint32_t max_buffer_size); /* Maximum allowable data buffer size */
   
/* De-initializes a channel structure.  Most likely called from client's softc
 * teardown function during driver detachment.  Will free any channel and buffer
 * resources */
void cx88_destroy_channel(cx88_channel *channel);

/* Synchronizes data buffer memory prior to transfer.
 * Returns 0 on success, appropriate error otherwise. */
int cx88_prepare_for_transfer(cx88_buffer *buffer);

/* Synchronizes data buffer memory following transfer */
void cx88_mark_transfer_complete(cx88_buffer *buffer);

/* Accessor for this device's board data structure.  Returns NULL
 * if device has not allocated board data or retrieved it from a sibling. */
cx88_board_data *cx88_common_get_board_data(device_t dev);

/* Defines an accessor function for generic shared data allocated by sibing */
typedef void *(*get_board_data_func) (device_t);

/* Adds a data structure to be shared by sibling devices.  Will attempt to
 * retrieve the structure from siblings using the provided accessor function,
 * and will allocate the structure if all siblings return NULL.
 * In spite of the name, this (along w/ cx88_remove_board_data()) is
 * designed to support not just the cx88_board_data structure used by the
 * common layer, but any arbitrary data a driver may want to share with its
 * siblings.
 * Returns 0 on success, appropriate error otherwise. */
int cx88_add_board_data(
   cx88_common_softc *sc,
   void **data,
   size_t data_size,
   get_board_data_func accessor,
   cx88_setup_func setup); /* Initialization function to be called if the shared data needs to be allocated */
   
/* Removes this driver's reference to a shared data structure.  If this driver is the last
 * driver to reference the structure, it will be freed.  Returns 0 on success, appropriate
 * error otherwise. */
int cx88_remove_board_data(
   cx88_common_softc *sc,
   void *data,
   cx88_teardown_func teardown); /* Will be called if the structure is to be freed */
   
/* Accessor for client softc.  If the softc is NULL or uninitialized (client's softc
 * setup callback has not been called), will return NULL */
void *cx88_get_softc(device_t dev);

/* Marks a device as busy.  Will return EBUSY if the device is already busy, 0 on success. */
int cx88_busy(cx88_common_softc *sc);

/* Marks a device as no longer busy */
void cx88_unbusy(cx88_common_softc *sc);

/* Write to GPIO pins: Each bit in mask selects the corresponding
 * GPIO pin for writing, and the corresponding bit in value is then written
 * to the pin */
void cx88_write_gpio(cx88_common_softc *sc, uint32_t mask, uint32_t value);

/* Read from GPIO pins: Each bit in mask selects the corresponding GPIO pin
 * for reading, and the corresponding bits read from the pins are returned */
uint32_t cx88_read_gpio(cx88_common_softc *sc, uint32_t mask);

/* Accesses card vendor data in PCI subsystem ID */
static inline uint32_t cx88_get_card(device_t dev)
{
   return (((uint32_t)(pci_get_subvendor(dev)) << 16) | pci_get_subdevice(dev));
}

/* Writes a cx88 register */
static inline void cx88_write_reg(
   cx88_common_softc *sc,
   uint32_t offset, /* Offset in cx88 memory window */
   uint32_t value)
{
   bus_space_write_4(sc->mem_tag, sc->mem_handle, offset, value);
}

/* Writes a contiguous series of registers */
static inline void cx88_write_regs(
   cx88_common_softc *sc,
   uint32_t offset,
   uint32_t *values,
   uint32_t count)
{
   bus_space_write_region_4(sc->mem_tag, sc->mem_handle, offset, values, count);
}

/* Reads a contiguous series of registers */
static inline void cx88_read_regs(
   cx88_common_softc *sc,
   uint32_t offset,
   uint32_t *values,
   uint32_t count)
{
   bus_space_read_region_4(sc->mem_tag, sc->mem_handle, offset, values, count);  
}

/* Reads a cx88 register */
static inline uint32_t cx88_read_reg(
   cx88_common_softc *sc,
   uint32_t offset) /* Offset in cx88 memory window */
{
   return bus_space_read_4(sc->mem_tag, sc->mem_handle, offset);    
}

/* Generates a bitfield */
static inline uint32_t cx88_write_field(
   uint32_t mask,
   uint8_t shift,
   uint32_t value)
{
   return ((value & mask) << shift);
}

/* Extracts a bitfield */
static inline uint32_t cx88_read_field(
   uint32_t value,
   uint32_t mask,
   uint8_t shift)
{
   return ((value & (mask << shift)) >> shift);
}

/* Acquires exclusive access to data that may be shared by sibling devices */
static inline void cx88_lock_board(cx88_common_softc *sc)
{
   mtx_lock(&(sc->board_data->lock));  
}

/* Releases exclusive access to data that may be shared by sibling devices */
static inline void cx88_unlock_board(cx88_common_softc *sc)
{
   mtx_unlock(&(sc->board_data->lock));
}

/* DEVMETHODs all client drivers should implement */
#define CX88_METHODS(reconfig_request) \
   DEVMETHOD(cx88_get_board_data,      cx88_common_get_board_data), \
   DEVMETHOD(cx88_request_reconfig,    reconfig_request)

/* All client drivers should declare dependency on this module */
#define CX88_DRIVER_MODULE(host) \
   MODULE_DEPEND(host, cx88, CX88_MINVER, CX88_MODVER, CX88_MAXVER);

#endif
