/*-
 * Copyright 2007 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

 /* File: cx88_risc.h
  * Purpose: Header file for cx88 RISC program generator functions
  */

#ifndef __CX88_RISC_H__
#define __CX88_RISC_H__

#include <sys/types.h>
#include "cx88_common.h"

/* Determines which (if any) RISC interrupts are triggered when
 * a RISC instruction is executed */
typedef enum
{
   CX88_NO_TRIGGER = 0x0,
   CX88_TRIGGER_RISCI1, /* RISC interrupt 1 */
   CX88_TRIGGER_RISCI2, /* RISC interrupt 2 */
   CX88_TRIGGER_RISCI1_RISCI2 /* both RISC interrupts */
} cx88_risci_trigger;

/* Determines how the internal counter should be dealt with
 * when a RISC instruction is executed */
typedef enum
{
   CX88_IGNORE_COUNT = 0x0, /* No change */
   CX88_INCREMENT_COUNT = 0x1, /* Increment */
   CX88_RESET_COUNT = 0x3 /* Reset to 0 */
} cx88_count_control;

/* Size of each instruction (in 4-byte words) */
#define CX88_RISC_SYNC_SIZE         1
#define CX88_RISC_READ_WRITE_SIZE   2
#define CX88_RISC_READC_WRITEC_SIZE 1
#define CX88_RISC_JUMP_SIZE         2
#define CX88_RISC_SKIP_SIZE         1
#define CX88_RISC_WRITERM_SIZE      3
#define CX88_RISC_WRITECM_SIZE      3
#define CX88_RISC_WRITECR_SIZE      4

/* Will transfer as many bytes as possible with a single RISC
 * instruction, up to a masximum of count. Uses buffer's channel 
 * direction and line/DMA segment offsets to determine whether to
 * use READ/READC/WRITE/WRITEC, and returns the actual number of 
 * bytes that can be transferred. */
uint32_t cx88_risc_transfer(
   cx88_buffer *buffer,
   uint32_t count, /* Maximum number of bytes */
   cx88_risci_trigger intrs, /* Applies to each generated instruction */
   cx88_count_control countctl); /* Applies to each generated instruction */

/* Generates a SYNC instruction */
void cx88_risc_sync(
   cx88_buffer *buffer,
   uint16_t lineno,
   boolean_t resync, /* If TRUE, don't raise SYNC interrupt */
   cx88_risci_trigger intrs,
   cx88_count_control countctl);

/* Generates a jump instruction */
int cx88_risc_jump(
   cx88_buffer *src, /* Buffer RISC program from which to jump */
   cx88_buffer *dest, /* Buffer RISC program to jump into (may be same as src) */
   uint32_t offset, /* Offset (in 4-byte words) at which to start in dest */
   cx88_risci_trigger intrs,
   cx88_count_control countctl);

/* Jumps to next buffer in RISC program sequence */
static inline int cx88_risc_next_buffer(
   cx88_buffer *buffer,
   uint32_t offset,
   cx88_risci_trigger intrs,
   cx88_count_control countctl)
{
   return (cx88_risc_jump(buffer, buffer->next_buffer, offset, intrs, countctl));
}

/* Ignores as many bytes as possible with a a single SKIP instruction,
 * up to a maximum of count bytes.  Returns the actual number of bytes that
 * can be skipped. */
uint32_t cx88_risc_skip(
   cx88_buffer *buffer,
   uint32_t count, /* Maximum number of bytes */
   cx88_risci_trigger intrs,
   cx88_count_control countctl);

/* Write cx88 register contents to memory */
void cx88_risc_write_reg_to_mem(
   cx88_buffer *buffer,
   uint32_t reg, /* Offset into cx88 memory window */
   uint32_t target_addr, /* Target physical address */
   cx88_risci_trigger intrs,
   cx88_count_control countctl);

/* Write constant to memory */
void cx88_risc_write_const_to_mem(
   cx88_buffer *buffer,
   uint32_t constant,
   uint32_t target_addr, /* Target physical address */
   cx88_risci_trigger intrs,
   cx88_count_control countctl);

/* Write constant to cx88 register */
void cx88_risc_write_const_to_reg(
   cx88_buffer *buffer,
   uint32_t constant,
   uint32_t mask, /* Bitmask to apply to constant */
   uint32_t reg, /* Offset into cx88 memory window */
   cx88_risci_trigger intrs,
   cx88_count_control countctl);

/* Generic function for transferring an arbitrary number of bytes.
 * Calls cx88_risc_transfer() as many times as necessary to generate the
 * instructions to transfer count bytes */
int cx88_risc_transfer_dmabuf (
   cx88_buffer *buffer,
   uint32_t count, /* Number of bytes */
   cx88_risci_trigger intrs,
   cx88_count_control countctl);
   
/* Generic function for ignoring an arbitrary number of bytes.
 * Calls cx88_risc_skip() as many times as necessary to generate the 
 * required SKIP instructions. */
int cx88_risc_skip_dmabuf(
   cx88_buffer *buffer,
   uint32_t count, /* Number of bytes */
   cx88_risci_trigger intrs,
   cx88_count_control countctl);

/* Determines whether the provided number of packets can fit into a whole
 * number of lines.  Useful for continuous transfers to ensure that the 
 * line position (SOL/EOL) of the first instruction will always be correct
 * when the RISC program jumps back to the beginning.  Unnecessary for 
 * line-based channels (line_size == packet_size). */
uint32_t cx88_risc_check_bufsize(cx88_channel *channel, uint32_t num_packets);

#endif
