/*-
 * Copyright 2007 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/bus.h>
#include <sys/conf.h>
#include <sys/lock.h>
#include <sys/sx.h>
#include <sys/module.h>
#include <sys/mman.h>
#include <sys/time.h>
#include <sys/selinfo.h>
#include <sys/poll.h>
#include <sys/fcntl.h>
#include "cx88_common.h"
#include "cx88_risc.h"
#include "cx88_if.h"
#include "cx88_i2c.h"
#include "cx88_i2c_if.h"
#include "cx88_regs.h"

#define MAX_VIDEO_BUFFER_SIZE (1 << 20)

typedef struct
{
   cx88_common_softc common;
   cx88_channel ch_packed;
   cx88_channel ch_u;
   cx88_channel ch_v;
   cx88_channel ch_vbi;
   struct mtx isrlock;
   boolean_t data_avail;
   struct sx sxlock;
   struct selinfo selectors;
   cx88_i2c_bus i2c;
   cx88_i2c_dev demod;
   cx88_i2c_dev tuner;
} cx88_video_softc;

static int cx88_video_probe(device_t dev);
static int cx88_video_attach(device_t dev);
static int cx88_video_setup (void *arg);
static int cx88_video_teardown (void *arg);
static void cx88_video_intr (void *arg);
static cx88_i2c_bus *cx88_video_get_i2c_bus(device_t dev);
static cx88_i2c_dev *cx88_video_get_i2c_dev(device_t dev, uint8_t addr);
static int cx88_video_request_reconfig(device_t dev);

static const cx88_client_driver cx88_video_client = 
{
   .softc_size               = sizeof(cx88_video_softc),
   .function                 = CX88_FUNC_VIDEO,
   .channel_mask             = CX88_CHANNEL_MASK(CX88_CH_VIDEO_PACKED) 
                                 | CX88_CHANNEL_MASK(CX88_CH_VIDEO_U)
                                 | CX88_CHANNEL_MASK(CX88_CH_VIDEO_V)
                                 | CX88_CHANNEL_MASK(CX88_CH_VIDEO_VBI),
   .setup                    = cx88_video_setup,
   .teardown                 = cx88_video_teardown,
   .interrupt_handler        = cx88_video_intr,
   .description              = "Conexant CX2388x Analog Video"
};

static devclass_t cx88_video_devclass;

/*static d_open_t cx88_video_open;
static d_close_t cx88_video_close;
static d_mmap_t cx88_video_mmap;
static d_ioctl_t cx88_video_ioctl;
static d_poll_t cx88_video_poll;*/

//static struct cdevsw cx88_video_cdevsw =
//{
   //.d_version = D_VERSION,
   //.d_flags = D_TRACKCLOSE,
   //.d_name = "cx88video"//,
   //.d_open = cx88_video_open,   /* open */
   //.d_close = cx88_video_close, /* close */
   //.d_mmap = cx88_video_mmap,   /* mmap */
   //.d_ioctl = cx88_video_ioctl, /* ioctl */
   //.d_poll = cx88_video_poll    /* poll */
//};

static int cx88_video_setup (void *arg)
{   
   cx88_video_softc *sc = arg;
   int error = 0;
   mtx_init(&(sc->isrlock), "CX88 Video ISR lock", NULL, MTX_DEF);
   sx_init(&(sc->sxlock), "CX88 Video SX lock");
   //int unit = device_get_unit(sc->common.dev);
   if ((error = cx88_init_channel(&(sc->ch_packed), &(sc->common), CX88_CH_VIDEO_PACKED, MAX_VIDEO_BUFFER_SIZE)))
   {
      goto fail0;
   }
   if ((error = cx88_init_channel(&(sc->ch_u), &(sc->common), CX88_CH_VIDEO_U, MAX_VIDEO_BUFFER_SIZE)))
   {
      goto fail1;
   }
   if ((error = cx88_init_channel(&(sc->ch_v), &(sc->common), CX88_CH_VIDEO_V, MAX_VIDEO_BUFFER_SIZE)))
   {
      goto fail2;
   }
   if ((error = cx88_init_channel(&(sc->ch_vbi), &(sc->common), CX88_CH_VIDEO_VBI, MAX_VIDEO_BUFFER_SIZE)))
   {
      goto fail3;
   }
   if ((error = cx88_i2c_create(&(sc->i2c), &(sc->common))))
   {
      goto fail4;
   }
   goto success;
   fail4:
      cx88_destroy_channel (&(sc->ch_vbi));
   fail3:
      cx88_destroy_channel (&(sc->ch_v));
   fail2:
      cx88_destroy_channel (&(sc->ch_u));
   fail1:
      cx88_destroy_channel (&(sc->ch_packed));
   fail0:
      mtx_destroy(&(sc->isrlock));
      sx_destroy(&(sc->sxlock));
   success:
      return error;
}

static int cx88_video_teardown (void *arg)
{
   int error = 0;
   cx88_video_softc *sc = arg;
   error = cx88_i2c_destroy(&(sc->i2c));
   if (!error)
   {
      cx88_destroy_channel (&(sc->ch_vbi));
      cx88_destroy_channel (&(sc->ch_v));
      cx88_destroy_channel (&(sc->ch_u));
      cx88_destroy_channel (&(sc->ch_packed));
      mtx_destroy(&(sc->isrlock));
      sx_destroy(&(sc->sxlock));
   }
   return error;
}

static void cx88_video_intr (void *arg)
{
}

static int cx88_video_probe(device_t dev) 
{
   return cx88_common_probe(dev, &cx88_video_client);
}

static int cx88_video_attach(device_t dev) 
{
   return cx88_common_attach(dev, &cx88_video_client);
}

static cx88_i2c_bus *cx88_video_get_i2c_bus(device_t dev)
{
   cx88_video_softc *sc = cx88_get_softc(dev);
   if (sc == NULL)
   {
      return NULL;
   }
   return (&(sc->i2c));
}

static cx88_i2c_dev *cx88_video_get_i2c_dev(device_t dev, uint8_t addr)
{
   return NULL;
}

static int cx88_video_request_reconfig(device_t dev)
{
   return ENODEV;   
}

static device_method_t cx88_video_methods[] = 
{
   /* Device interface */
   DEVMETHOD(device_probe,             cx88_video_probe),
   DEVMETHOD(device_attach,            cx88_video_attach),
   DEVMETHOD(device_detach,            cx88_common_detach),
   DEVMETHOD(device_shutdown,          cx88_common_detach),
   DEVMETHOD(device_suspend,           cx88_common_detach),
   DEVMETHOD(device_resume,            cx88_video_attach),

   /* cx88 interface */
   CX88_METHODS(cx88_video_request_reconfig),

   /* i2c host methods */
   CX88_I2C_METHODS(cx88_video_get_i2c_bus, cx88_video_get_i2c_dev),

   {NULL, NULL}
};

static driver_t cx88_video_driver =
{
   "cx88video",
   cx88_video_methods,
   0
};

DRIVER_MODULE(cx88video, pci, cx88_video_driver, cx88_video_devclass, NULL, NULL);
CX88_DRIVER_MODULE(cx88video);
CX88_I2C_DRIVER_MODULE(cx88video);
MODULE_VERSION(cx88video, 1);
